from PyQt5 import QtCore, QtGui, QtWidgets

DEFAULT_CSS = """
QRangeSlider * {
    border: 0px;
    padding: 0px;
}
QRangeSlider #Head {
    background: #cef;
}
QRangeSlider #Span {
    background: #16f;
}
QRangeSlider #Span:active {
    background: #05f;
}
QRangeSlider #Tail {
    background: #cef;
}
QRangeSlider > QSplitter::handle {
    background: #08f;
}
QRangeSlider > QSplitter::handle:vertical {
    height: 3px;
}
QRangeSlider > QSplitter::handle:pressed {
    background: #0cf;
}
"""


def scale(val, src, dst):
    ret = ((val - src[0]) / float(src[1]-src[0])) * (dst[1]-dst[0]) + dst[0]
    return int(ret)


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("QRangeSlider")
        Form.setStyleSheet(DEFAULT_CSS)
        self.gridLayout = QtWidgets.QGridLayout(Form)
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.gridLayout.setSpacing(0)
        self.gridLayout.setObjectName("gridLayout")
        self._splitter = QtWidgets.QSplitter(Form)
        self._splitter.setMinimumSize(QtCore.QSize(0, 0))
        self._splitter.setMaximumSize(QtCore.QSize(16777215, 16777215))
        self._splitter.setOrientation(QtCore.Qt.Horizontal)
        self._splitter.setObjectName("splitter")
        self._head = QtWidgets.QGroupBox(self._splitter)
        self._head.setTitle("")
        self._head.setObjectName("Head")
        self._handle = QtWidgets.QGroupBox(self._splitter)
        self._handle.setTitle("")
        self._handle.setObjectName("Span")
        self._tail = QtWidgets.QGroupBox(self._splitter)
        self._tail.setTitle("")
        self._tail.setObjectName("Tail")
        self.gridLayout.addWidget(self._splitter, 0, 0, 1, 1)
        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("QRangeSlider", "QRangeSlider"))


class Element(QtWidgets.QGroupBox):
    def __init__(self, parent, main):
        super(Element, self).__init__(parent)
        self.main = main

    def setStyleSheet(self, style):
        self.parent().setStyleSheet(style)

    def textColor(self):
        return getattr(self, '__textColor', QtGui.QColor(125, 125, 125))

    def setTextColor(self, color):
        if type(color) == tuple and len(color) == 3:
            color = QtGui.QColor(color[0], color[1], color[2])
        elif type(color) == int:
            color = QtGui.QColor(color, color, color)
        setattr(self, '__textColor', color)

    def paintEvent(self, event):
        qp = QtGui.QPainter()
        qp.begin(self)
        if self.main.drawValues():
            self.drawText(event, qp)
        qp.end()


class Head(Element):
    def __init__(self, parent, main):
        super(Head, self).__init__(parent, main)

    def drawText(self, event, qp):
        qp.setPen(self.textColor())
        qp.setFont(QtGui.QFont('Arial', 10))
        qp.drawText(event.rect(), QtCore.Qt.AlignLeft, str(self.main.min()))


class Tail(Element):
    def __init__(self, parent, main):
        super(Tail, self).__init__(parent, main)

    def drawText(self, event, qp):
        qp.setPen(self.textColor())
        qp.setFont(QtGui.QFont('Arial', 10))
        qp.drawText(event.rect(), QtCore.Qt.AlignRight, str(self.main.max()))


class Handle(Element):
    def __init__(self, parent, main):
        super(Handle, self).__init__(parent, main)

    def drawText(self, event, qp):
        qp.setPen(self.textColor())
        qp.setFont(QtGui.QFont('Arial', 10))
        qp.drawText(event.rect(), QtCore.Qt.AlignLeft, str(self.main.start()))
        qp.drawText(event.rect(), QtCore.Qt.AlignRight, str(self.main.end()))


class QRangeSlider(QtWidgets.QWidget, Ui_Form):
    minValueChanged = QtCore.pyqtSignal(int)
    maxValueChanged = QtCore.pyqtSignal(int)
    startValueChanged = QtCore.pyqtSignal(int)
    endValueChanged = QtCore.pyqtSignal(int)
    rangeChanged = QtCore.pyqtSignal(int, int)

    _INT_NUM = 2000
    _SPLIT_START = 1
    _SPLIT_END = 2

    def __init__(self, parent=None):
        super(QRangeSlider, self).__init__(parent)
        self.setupUi(self)
        self.setFixedHeight(21)
        self.setMouseTracking(False)
        self._splitter.splitterMoved.connect(self._handleMoveSplitter)
        self._head_layout = QtWidgets.QHBoxLayout()
        self._head_layout.setSpacing(0)
        self._head_layout.setContentsMargins(0, 0, 0, 0)
        self._head.setLayout(self._head_layout)
        self.head = Head(self._head, main=self)
        self._head_layout.addWidget(self.head)
        self._handle_layout = QtWidgets.QHBoxLayout()
        self._handle_layout.setSpacing(0)
        self._handle_layout.setContentsMargins(0, 0, 0, 0)
        self._handle.setLayout(self._handle_layout)
        self.handle = Handle(self._handle, main=self)
        self._handle_layout.addWidget(self.handle)
        self._tail_layout = QtWidgets.QHBoxLayout()
        self._tail_layout.setSpacing(0)
        self._tail_layout.setContentsMargins(0, 0, 0, 0)
        self._tail.setLayout(self._tail_layout)
        self.tail = Tail(self._tail, main=self)
        self._tail_layout.addWidget(self.tail)
        # set limits
        self.setMin(0)
        self.setMax(self._INT_NUM)
        self.setStart(0)
        self.setEnd(self._INT_NUM)
        self.setDrawValues(False)

    def min(self):
        return getattr(self, '__min', None)

    def max(self):
        return getattr(self, '__max', None)

    def setMin(self, value):
        setattr(self, '__min', value)
        self.minValueChanged.emit(value)

    def setMax(self, value):
        setattr(self, '__max', value)
        self.maxValueChanged.emit(value)

    def start(self):
        return getattr(self, '__start', None)

    def end(self):
        return getattr(self, '__end', None)

    def _setStart(self, value):
        setattr(self, '__start', value)
        self.startValueChanged.emit(value)
        self.rangeChanged.emit(value, self.end())

    def setStart(self, value):
        v = self._valueToPos(value)
        self._splitter.splitterMoved.disconnect()
        self._splitter.moveSplitter(v, self._SPLIT_START)
        self._splitter.refresh()
        self._splitter.splitterMoved.connect(self._handleMoveSplitter)
        self._setStart(value)

    def _setEnd(self, value):
        setattr(self, '__end', value)
        self.endValueChanged.emit(value)
        self.rangeChanged.emit(self.start(), value)

    def setEnd(self, value):
        v = self._valueToPos(value)
        self._splitter.splitterMoved.disconnect()
        self._splitter.moveSplitter(v, self._SPLIT_END)
        self._splitter.refresh()
        self._splitter.splitterMoved.connect(self._handleMoveSplitter)
        self._setEnd(value)

    def drawValues(self):
        return getattr(self, '__drawValues', None)

    def setDrawValues(self, draw):
        setattr(self, '__drawValues', draw)

    def getRange(self):
        return (self.start(), self.end())

    def setRange(self, start, end):
        self.setStart(start)
        self.setEnd(end)

    def setBackgroundStyle(self, style):
        self._tail.setStyleSheet(style)
        self._head.setStyleSheet(style)

    def setSpanStyle(self, style):
        self._handle.setStyleSheet(style)

    def _valueToPos(self, value):
        return scale(value, (self.min(), self.max()), (0, self.width()))

    def _posToValue(self, xpos):
        return scale(xpos, (0, self.width()), (self.min(), self.max()))

    def _handleMoveSplitter(self, xpos, index):
        hw = self._splitter.handleWidth()
        if index == self._SPLIT_START:
            v = self._posToValue(xpos)
            if v >= self.end():
                self.setEnd(self._INT_NUM)
            self._setStart(v)
        elif index == self._SPLIT_END:
            v = self._posToValue(xpos + hw)
            if v <= self.start():
                self.setStart(0)
            self._setEnd(v)



import pkg_resources

from PyQt5 import uic, QtWidgets, QtCore, QtGui

from ... import meta_tool
from ... import pipeline


class MatrixDataset(QtWidgets.QWidget):
    active_toggled = QtCore.pyqtSignal()
    enabled_toggled = QtCore.pyqtSignal(bool)
    option_action = QtCore.pyqtSignal(str)
    modify_clicked = QtCore.pyqtSignal(str)

    def __init__(self, identifier=None, state=None):
        """Create a new dataset matrix element

        Specify either an existing Dataslot identifier or a
        Dataslot state
        """
        QtWidgets.QWidget.__init__(self)
        path_ui = pkg_resources.resource_filename(
            "shapeout2.gui.matrix", "dm_dataset.ui")
        uic.loadUi(path_ui, self)

        # options button
        menu = QtWidgets.QMenu()
        menu.addAction('insert anew', self.action_insert_anew)
        menu.addAction('duplicate', self.action_duplicate)
        menu.addAction('remove', self.action_remove)
        self.toolButton_opt.setMenu(menu)

        # toggle all active, all inactive, semi state
        self.toolButton_toggle.clicked.connect(self.active_toggled.emit)

        # toggle enabled/disabled state
        self.checkBox.clicked.connect(self.enabled_toggled.emit)

        # modify slot button
        self.toolButton_modify.clicked.connect(self.on_modify)

        if state is None:
            slot = pipeline.Dataslot._instances[identifier]
            self.identifier = identifier
            self.path = slot.path
            # set tooltip/label
            self.update_content()
        else:
            self.__setstate__(state)

    def __getstate__(self):
        state = {"path": self.path,
                 "identifier": self.identifier,
                 "enabled": self.checkBox.isChecked(),
                 }
        return state

    def __setstate__(self, state):
        self.identifier = state["identifier"]
        self.path = state["path"]
        self.checkBox.setChecked(state["enabled"])
        self.update_content()

    def action_duplicate(self):
        self.option_action.emit("duplicate")

    def action_insert_anew(self):
        self.option_action.emit("insert_anew")

    def action_remove(self):
        self.option_action.emit("remove")

    def on_modify(self):
        self.modify_clicked.emit(self.identifier)

    def set_label_string(self, string):
        if self.label.fontMetrics().boundingRect(string).width() < 65:
            nstring = string
        else:
            nstring = string + "..."
            while True:
                width = self.label.fontMetrics().boundingRect(nstring).width()
                if width > 65:
                    nstring = nstring[:-4] + "..."
                else:
                    break
        self.label.setText(nstring)

    def update_content(self):
        """Reset tool tips and title"""
        if self.path is not None:
            tip = meta_tool.get_repr(self.path, append_path=True)
            self.setToolTip(tip)
            self.label.setToolTip(tip)
            slot = pipeline.Dataslot._instances[self.identifier]
            name = slot.name
            self.set_label_string(name)
            # Set region image
            region = meta_tool.get_info(self.path,
                                        section="setup",
                                        key="chip region")
            icon = QtGui.QIcon.fromTheme("region_{}".format(region))
            pixmap = icon.pixmap(16)
            self.label_region.setPixmap(pixmap)
            self.label_region.setToolTip(region)
            if region == "channel":
                # Set flow rate
                flow_rate = meta_tool.get_info(self.path,
                                               section="setup",
                                               key="flow rate")
                self.label_flowrate.setText("{:.4g}".format(flow_rate))
                self.label_flowrate.setToolTip("{:.4g} µL/s".format(flow_rate))
            else:
                self.label_flowrate.setText(region[:3])
                self.label_flowrate.setToolTip(region)



from PyQt5 import QtCore, QtWidgets


class MDISubWindowWOButtons(QtWidgets.QMdiSubWindow):
    def __init__(self, *args, **kwargs):
        super(MDISubWindowWOButtons, self).__init__(*args, **kwargs)
        self.setSystemMenu(None)
        self.setWindowFlags(QtCore.Qt.CustomizeWindowHint
                            | QtCore.Qt.WindowTitleHint)

    def closeEvent(self, event):
        event.ignore()



import numpy as np
from PyQt5 import QtWidgets, QtGui


class DoubleSpinBoxNan(QtWidgets.QDoubleSpinBox):
    """A doubleSpinBox that uses minimum() and maximum() as np.nan"""

    def __init__(self, *args, **kwargs):
        super(DoubleSpinBoxNan, self).__init__(*args, **kwargs)
        self._suffix = self.suffix()  # remember initial suffix
        self.validator = NanFloatValidator()

    def validate(self, text, position):
        return self.validator.validate(text, position, self.suffix())

    def value(self):
        value = super(DoubleSpinBoxNan, self).value()
        if value == self.minimum() or value == self.maximum():
            value = np.nan
        return value

    def valueFromText(self, text):
        if text == "nan":
            return np.nan
        else:
            return convert_string_to_nanfloat(text[:-len(self.suffix())])

    def textFromValue(self, value):
        if value == self.minimum() or value == self.maximum():
            return "nan"
        else:
            return str(value)


class NanFloatValidator(QtGui.QValidator):
    def validate(self, text, position, suffix):
        string = text[:-len(suffix)]
        if string in ["n", "na", "nan"]:
            text = "nan"
        if valid_nanfloat_string(string):
            return self.Acceptable, text, position
        if string == "":
            return self.Intermediate, text, position
        return self.Invalid, text, position

    def fixup(self, text):
        try:
            val = convert_string_to_nanfloat(text)
        except ValueError:
            val = ""
        return str(val)


def valid_nanfloat_string(string):
    try:
        convert_string_to_nanfloat(string)
    except ValueError:
        valid = False
    else:
        valid = True
    return valid


def convert_string_to_nanfloat(string):
    try:
        val = float(string)
    except ValueError:
        string = string.strip()
        for iid in ["n", "na", "nan"]:
            if string.startswith(iid):
                val = np.nan
                break
        else:
            raise ValueError("Not a valid nan-float!")
    return val



import pathlib

import numpy as np
from shapeout2 import pipeline


def test_get_heredity():
    path = pathlib.Path(__file__).parent / "data" / "calibration_beads_47.rtdc"

    # initialize
    slot = pipeline.Dataslot(path)
    ds = slot.get_dataset()
    ray = pipeline.FilterRay(slot)

    # come up with a few filters
    # this does some simple things
    filt1 = pipeline.Filter()
    filt1.boxdict["area_um"] = {"start": ds["area_um"].min(),
                                "end": np.mean(ds["area_um"]),
                                "active": True}
    # this one does nothing (and should be ignored)
    filt2 = pipeline.Filter()
    filt2.filter_used = False
    # another one with simple things
    filt3 = pipeline.Filter()
    filt3.boxdict["deform"] = {"start": ds["deform"].min(),
                               "end": np.mean(ds["deform"]),
                               "active": True}

    # simple
    ds1 = ray.get_dataset(filters=[filt1, filt2], apply_filter=True)
    assert ray._generation == 0

    # leaving out a filter that does nothing will not change anything
    ds2 = ray.get_dataset(filters=[filt1], apply_filter=True)
    assert ray._generation == 0
    assert ds1 is ds2

    # changing the order will not change anything either
    ds3 = ray.get_dataset(filters=[filt2, filt1], apply_filter=True)
    assert ray._generation == 0
    assert ds1 is ds3

    # adding a new filter will change the dataset
    ds4 = ray.get_dataset(filters=[filt1, filt2, filt3], apply_filter=True)
    assert ray._generation == 0
    assert ds1 is not ds4
    assert ds1 is ds4.hparent

    # going back does not increment the generation...
    ds5 = ray.get_dataset(filters=[filt1, filt2], apply_filter=True)
    assert ray._generation == 0
    assert ds1 is ds5

    # ...but changing the order is
    ds5 = ray.get_dataset(filters=[filt3, filt1, filt2], apply_filter=True)
    assert ray._generation == 1
    assert ds1 is not ds5
    assert ds3 is not ds5

    # and then again, when we remove a filter, we get something different
    ds6 = ray.get_dataset(filters=[filt3, filt1], apply_filter=True)
    assert ray._generation == 1
    assert ds1 is not ds6
    assert ds5 is ds6  # b/c filt2 does nothing


def test_filtering():
    path = pathlib.Path(__file__).parent / "data" / "calibration_beads_47.rtdc"

    # initialize
    slot = pipeline.Dataslot(path)
    ds = slot.get_dataset()
    ray = pipeline.FilterRay(slot)

    # come up with a few filters
    # this does some simple things
    filt1 = pipeline.Filter()
    filt1.boxdict["area_um"] = {"start": ds["area_um"].min(),
                                "end": np.mean(ds["area_um"]),
                                "active": True}
    # this one does nothing (and should be ignored)
    filt2 = pipeline.Filter()
    filt2.filter_used = False
    # another one with simple things
    filt3 = pipeline.Filter()
    filt3.boxdict["deform"] = {"start": ds["deform"].min(),
                               "end": np.mean(ds["deform"]),
                               "active": True}

    ds2 = ray.get_dataset(filters=[filt2, filt1], apply_filter=True)
    assert len(ds2) == 47
    ds3 = ray.get_dataset(filters=[filt1, filt3], apply_filter=True)
    assert len(ds3) == 22
    assert np.sum(ds2.filter.all) == 22
    assert np.sum(ds3.filter.all) == 12
    ds1 = ray.get_dataset(filters=[filt2], apply_filter=True)
    assert np.sum(ds1.filter.all) == len(ds)
    ds4 = ray.get_dataset(filters=[filt1, filt2], apply_filter=True)
    assert len(ds4) == 47


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



from .block_matrix import BlockMatrix  # noqa: F401
from .dm_dataset import MatrixDataset  # noqa: F401
from .dm_element import MatrixElement  # noqa: F401
from .dm_filter import MatrixFilter  # noqa: F401



import copy
import pkg_resources
import warnings

import dclab
from dclab import kde_contours
import numpy as np
from PyQt5 import uic, QtCore, QtGui, QtWidgets
import pyqtgraph as pg
from pyqtgraph import exporters
from pyqtgraph.graphicsItems.GradientEditorItem import Gradients


from .. import plot_cache
from .. import util

from .widgets import ColorBarWidget, SimplePlotItem


# Register custom colormaps
Gradients["grayblue"] = {'ticks': [(0.0, (100, 100, 100, 255)),
                                   (1.0, (0, 0, 255, 255))],
                         'mode': 'rgb'}

Gradients["graygreen"] = {'ticks': [(0.0, (100, 100, 100, 255)),
                                    (1.0, (0, 180, 0, 255))],
                          'mode': 'rgb'}

Gradients["grayorange"] = {'ticks': [(0.0, (100, 100, 100, 255)),
                                     (1.0, (210, 110, 0, 255))],
                           'mode': 'rgb'}

Gradients["grayred"] = {'ticks': [(0.0, (100, 100, 100, 255)),
                                  (1.0, (200, 0, 0, 255))],
                        'mode': 'rgb'}


class ContourSpacingTooLarge(UserWarning):
    pass


class PipelinePlot(QtWidgets.QWidget):
    """Implements the plotting pipeline using pyqtgraph"""
    instances = {}

    def __init__(self, parent, pipeline, plot_id, *args, **kwargs):
        QtWidgets.QWidget.__init__(self, parent=parent, *args, **kwargs)
        path_ui = pkg_resources.resource_filename(
            "shapeout2.gui", "pipeline_plot.ui")
        uic.loadUi(path_ui, self)
        # used to avoid unnecessary plotting
        self._plot_data_hash = "unset"

        self.pipeline = pipeline
        self.identifier = plot_id
        self.update_content()
        PipelinePlot.instances[plot_id] = self

    def update_content(self):
        """Update the current plot"""
        parent = self.parent()
        dslist, slot_states = self.pipeline.get_plot_datasets(self.identifier)
        plot = self.pipeline.get_plot(self.identifier)
        plot_state = plot.__getstate__()
        # check whether anything changed
        # 1. plot state and all relevant slot states
        tohash = [slot_states, plot_state]
        # 2. all relevant filter states
        for slot_state in slot_states:
            slot_id = slot_state["identifier"]
            for filt_id in self.pipeline.filter_ids:
                if self.pipeline.is_element_active(slot_id, filt_id):
                    filt = self.pipeline.get_filter(filt_id)
                    filt_state = filt.__getstate__()
                    tohash.append([slot_id, filt_id, filt_state])
                    # also check whether the polygon filters changed (#26)
                    for pid in filt_state["polygon filters"]:
                        pf = dclab.PolygonFilter.get_instance_from_id(pid)
                        tohash.append(pf.__getstate__())
        plot_data_hash = util.hashobj(tohash)
        if plot_data_hash == self._plot_data_hash:
            # do nothing
            return
        else:
            self._plot_data_hash = plot_data_hash

        # abbreviations
        gen = plot_state["general"]
        lay = plot_state["layout"]
        sca = plot_state["scatter"]

        # auto range (overrides stored ranges)
        if gen["auto range"]:
            # default range is limits + 5% margin
            gen["range x"] = self.pipeline.get_min_max(feat=gen["axis x"],
                                                       plot_id=self.identifier,
                                                       margin=.05)
            gen["range y"] = self.pipeline.get_min_max(feat=gen["axis y"],
                                                       plot_id=self.identifier,
                                                       margin=0.05)

        # title
        self.setWindowTitle(lay["name"])

        # clear widget
        self.plot_layout.clear()

        if not slot_states:
            return

        labelx, labely = get_axes_labels(plot_state, slot_states)

        # font size for plot title (default size + 2)
        size = "{}pt".format(QtGui.QFont().pointSize() + 2)
        self.plot_layout.addLabel(lay["name"], colspan=3, size=size)
        self.plot_layout.nextRow()

        self.plot_layout.addLabel(labely, angle=-90)
        linner = self.plot_layout.addLayout()

        # limits in case of scatter plot and feature hue
        if lay["division"] == "merge":
            pp = PipelinePlotItem(parent=linner)
            linner.addItem(item=pp,
                           row=None,
                           col=None,
                           rowspan=1,
                           colspan=1)
            pp.redraw(dslist, slot_states, plot_state)
        elif lay["division"] == "each":
            colcount = 0
            for ds, sl in zip(dslist, slot_states):
                pp = PipelinePlotItem(parent=linner)
                linner.addItem(item=pp,
                               row=None,
                               col=None,
                               rowspan=1,
                               colspan=1)
                pp.redraw([ds], [sl], plot_state)
                colcount += 1
                if colcount % lay["column count"] == 0:
                    linner.nextRow()
        elif lay["division"] == "multiscatter+contour":
            colcount = 0
            # scatter plots
            plot_state_scatter = copy.deepcopy(plot_state)
            plot_state_scatter["contour"]["enabled"] = False
            for ds, sl in zip(dslist, slot_states):
                pp = PipelinePlotItem(parent=linner)
                linner.addItem(item=pp,
                               row=None,
                               col=None,
                               rowspan=1,
                               colspan=1)
                pp.redraw([ds], [sl], plot_state_scatter)
                colcount += 1
                if colcount % lay["column count"] == 0:
                    linner.nextRow()
            # contour plot
            plot_state_contour = copy.deepcopy(plot_state)
            plot_state_contour["scatter"]["enabled"] = False
            pp = PipelinePlotItem(parent=linner)
            linner.addItem(item=pp,
                           row=None,
                           col=None,
                           rowspan=1,
                           colspan=1)
            pp.redraw(dslist, slot_states, plot_state_contour)

        # colorbar
        colorbar_kwds = {}

        if sca["marker hue"] == "kde":
            colorbar_kwds["vmin"] = 0
            colorbar_kwds["vmax"] = 1
            colorbar_kwds["label"] = "density [a.u.]"
        elif sca["marker hue"] == "feature":
            feat = sca["hue feature"]
            label = dclab.dfn.get_feature_label(feat)
            fl_names = slot_states[0]["fl names"]
            if label.count("FL"):
                for key in fl_names:
                    if key in label:
                        label = label.replace(key, fl_names[key])
                        break
            colorbar_kwds["label"] = label
            if label.endswith("[a.u.]"):
                colorbar_kwds["vmin"] = 0
                colorbar_kwds["vmax"] = 1
            else:
                colorbar_kwds["vmin"] = sca["hue min"]
                colorbar_kwds["vmax"] = sca["hue max"]

        if colorbar_kwds:
            # add colorbar
            colorbar = ColorBarWidget(
                cmap=sca["colormap"],
                width=15,
                height=min(300, lay["size y"]//2),
                **colorbar_kwds
            )
            self.plot_layout.addItem(colorbar)

        # x-axis label
        self.plot_layout.nextRow()
        self.plot_layout.addLabel(labelx, col=1)

        # Set size in the end (after layout is populated)
        self.setMinimumSize(lay["size x"], lay["size y"])
        self.setMaximumSize(lay["size x"], lay["size y"])
        size_hint = self.parent().sizeHint()
        parent.setMinimumSize(size_hint)
        parent.setMaximumSize(size_hint)


class PipelinePlotItem(SimplePlotItem):
    def __init__(self, *args, **kwargs):
        super(PipelinePlotItem, self).__init__(*args, **kwargs)
        # circumvent problems with removed plots
        self.setAcceptHoverEvents(False)
        # Disable user interaction
        self.setMouseEnabled(x=False, y=False)
        # bring axes to front
        self.axes_to_front()
        # Keep track of all elements (for redraw)
        self._plot_elements = []

    def perform_export(self, file):
        """Performs export in new layout with axes labels set

        Overrides the basic functionality of SimplePlotItem.
        See https://github.com/ZELLMECHANIK-DRESDEN/ShapeOut2/issues/7
        """
        # Create a plot window
        win = pg.GraphicsLayoutWidget(
            size=(self.width() + 100, self.height() + 100),
            show=True)
        # fill layout
        labelx, labely = get_axes_labels(self.plot_state, self.slot_states)
        win.addLabel(labely, angle=-90)
        explot = PipelinePlotItem()
        explot.redraw(self.dslist, self.slot_states, self.plot_state)
        win.addItem(explot)
        win.addLabel("")  # spacer to avoid cut tick labels on the right(#7)
        win.nextRow()
        win.addLabel(labelx, col=1)
        # Update the UI (do it twice, otherwise the tick labels overlap)
        QtWidgets.QApplication.processEvents()
        QtWidgets.QApplication.processEvents()
        win.hide()
        # perform actual export
        suffix = file[-3:]
        if suffix == "png":
            exp = exporters.ImageExporter(win.scene())
            # translate from screen resolution (80dpi) to 300dpi
            exp.params["width"] = int(exp.params["width"] / 72 * 300)
        elif suffix == "svg":
            exp = exporters.SVGExporter(win.scene())
        exp.export(file)

    def redraw(self, dslist, slot_states, plot_state):
        # Remove everything
        for el in self._plot_elements:
            self.removeItem(el)

        if not dslist:
            return

        self.dslist = dslist
        self.slot_states = slot_states
        self.plot_state = plot_state

        # General
        gen = plot_state["general"]
        # TODO:
        # - test whether all datasets have same channel width / pixel size
        # Isoelastics
        if gen["isoelastics"]:
            cfg = dslist[0].config
            els = add_isoelastics(plot_item=self,
                                  axis_x=gen["axis x"],
                                  axis_y=gen["axis y"],
                                  channel_width=cfg["setup"]["channel width"],
                                  pixel_size=cfg["imaging"]["pixel size"])
            self._plot_elements += els
        # Modifications in log mode
        set_viewbox(self,
                    range_x=gen["range x"],
                    range_y=gen["range y"],
                    scale_x=gen["scale x"],
                    scale_y=gen["scale y"])
        # Scatter data
        sca = plot_state["scatter"]
        if sca["enabled"]:
            for rtdc_ds, ss in zip(dslist, slot_states):
                sct = add_scatter(plot_item=self,
                                  rtdc_ds=rtdc_ds,
                                  plot_state=plot_state,
                                  slot_state=ss
                                  )
                self._plot_elements += sct
        # Contour data
        if plot_state["contour"]["enabled"]:
            # show legend
            if plot_state["contour"]["legend"]:
                legend = self.addLegend(offset=(-.01, +.01))
            else:
                legend = None
            for rtdc_ds, ss in zip(dslist, slot_states):
                con = add_contour(plot_item=self,
                                  rtdc_ds=rtdc_ds,
                                  plot_state=plot_state,
                                  slot_state=ss,
                                  legend=legend,
                                  )
                self._plot_elements += con

        # Set subplot title and number of events
        if plot_state["layout"]["label plots"]:
            if len(dslist) == 1 and plot_state["scatter"]["enabled"]:
                # only one scatter plot
                # set title
                ss = slot_states[0]
                thtmls = "<span style='color:{}'>{}</span>"
                title = thtmls.format(ss["color"], ss["name"])
                self.setTitle(title)
                if plot_state["scatter"]["show event count"]:
                    # set event count
                    chtml = "<span style='font-size:{}pt'>".format(
                        # default font size - 1
                        QtGui.QFont().pointSize() - 1) + "{} events</span>"
                    label = QtWidgets.QGraphicsTextItem(
                        "",
                        # This is kind of hackish: set the parent to the right
                        # axis so that it is always drawn there.
                        parent=self.axes["right"]["item"])
                    label.setHtml(chtml.format(len(sct[0].data)))
                    # move the label to the left by its width
                    label.setPos(-label.boundingRect().width()+2, -5)

            elif (plot_state["contour"]["enabled"]
                    and not plot_state["scatter"]["enabled"]):
                # only a contour plot
                # set title
                self.setTitle("Contours")


def add_contour(plot_item, plot_state, rtdc_ds, slot_state, legend=None):
    gen = plot_state["general"]
    con = plot_state["contour"]
    try:
        x, y, density = plot_cache.get_contour_data(
            rtdc_ds=rtdc_ds,
            xax=gen["axis x"],
            yax=gen["axis y"],
            xacc=con["spacing x"],
            yacc=con["spacing y"],
            xscale=gen["scale x"],
            yscale=gen["scale y"],
            kde_type=gen["kde"],
        )
    except ValueError:
        # most-likely there is nothing to compute a contour for
        return []
    if density.shape[0] < 3 or density.shape[1] < 3:
        warnings.warn("Contour not possible; spacing may be too large!",
                      ContourSpacingTooLarge)
        return []
    plev = kde_contours.get_quantile_levels(
        density=density,
        x=x,
        y=y,
        xp=rtdc_ds[gen["axis x"]][rtdc_ds.filter.all],
        yp=rtdc_ds[gen["axis y"]][rtdc_ds.filter.all],
        q=np.array(con["percentiles"])/100,
        normalize=True)
    contours = []
    for level in plev:
        # make sure that the contour levels are not at the boundaries
        if not (np.allclose(level, 0, atol=1e-12, rtol=0)
                or np.allclose(level, 1, atol=1e-12, rtol=0)):
            cc = kde_contours.find_contours_level(
                density, x=x, y=y, level=level)
            contours.append(cc)

    elements = []
    for ii in range(len(contours)):
        style = linestyles[con["line styles"][ii]]
        width = con["line widths"][ii]
        for cci in contours[ii]:
            cline = pg.PlotDataItem(x=cci[:, 0],
                                    y=cci[:, 1],
                                    pen=pg.mkPen(color=slot_state["color"],
                                                 width=width,
                                                 style=style,
                                                 ),
                                    )
            elements.append(cline)
            plot_item.addItem(cline)
            if ii == 0 and legend is not None:
                legend.addItem(cline, slot_state["name"])
            # Always plot higher percentiles above lower percentiles
            # (useful if there are multiple contour plots overlapping)
            cline.setZValue(con["percentiles"][ii])
    return elements


def add_isoelastics(plot_item, axis_x, axis_y, channel_width, pixel_size):
    elements = []
    isodef = dclab.isoelastics.get_default()
    # We do not use isodef.get_with_rtdcbase, because then the
    # isoelastics would be shifted according to flow rate and.
    # viscosity. We could do it, but for visualization there is
    # really no need and also, the plots then look the same as
    # in Shape-Out 1.
    try:
        iso = isodef.get(
            method="numerical",
            channel_width=channel_width,
            flow_rate=None,
            viscosity=None,
            col1=axis_x,
            col2=axis_y,
            add_px_err=True,
            px_um=pixel_size)
    except KeyError:
        pass
    else:
        for ss in iso:
            iline = pg.PlotDataItem(x=ss[:, 0], y=ss[:, 1])
            plot_item.addItem(iline)
            elements.append(iline)
            # send them to the back
            iline.setZValue(-100)
    return elements


def add_scatter(plot_item, plot_state, rtdc_ds, slot_state):
    gen = plot_state["general"]
    sca = plot_state["scatter"]
    scatter = pg.ScatterPlotItem(size=sca["marker size"],
                                 pen=pg.mkPen(color=(0, 0, 0, 0)),
                                 brush=pg.mkBrush("k"),
                                 symbol="s")
    scatter.setAcceptHoverEvents(False)
    plot_item.addItem(scatter)

    if sca["marker hue"] == "kde":
        kde_type = gen["kde"]
    else:
        kde_type = "none"

    x, y, kde, idx = plot_cache.get_scatter_data(
        rtdc_ds=rtdc_ds,
        downsample=sca["downsample"] * sca["downsampling value"],
        xax=gen["axis x"],
        yax=gen["axis y"],
        xscale=gen["scale x"],
        yscale=gen["scale y"],
        kde_type=kde_type,
    )
    # define colormap
    # TODO:
    # - common code base with QuickView
    cmap = pg.ColorMap(*zip(*Gradients[sca["colormap"]]["ticks"]))
    if sca["marker hue"] == "kde":
        brush = []
        # Note: we don't expand the density to [0, 1], because the
        # colorbar will show "density" and because we don want to
        # compute the density in this function and not someplace else.
        for k in kde:
            brush.append(cmap.mapToQColor(k))
        # Note, colors could also be digitized (does not seem to be faster):
        # cbin = np.linspace(0, 1, 1000)
        # dig = np.digitize(kde, cbin)
        # for idx in dig:
        #     brush.append(cmap.mapToQColor(cbin[idx]))
    elif sca["marker hue"] == "feature":
        brush = []
        feat = np.asarray(rtdc_ds[sca["hue feature"]][idx], dtype=float)
        feat -= sca["hue min"]
        feat /= sca["hue max"] - sca["hue min"]
        for f in feat:
            if np.isnan(f):
                brush.append(pg.mkColor("#FF0000"))
            else:
                brush.append(cmap.mapToQColor(f))
    elif sca["marker hue"] == "dataset":
        alpha = int(sca["marker alpha"] * 255)
        colord = pg.mkColor(slot_state["color"])
        colord.setAlpha(alpha)
        brush = pg.mkBrush(colord)
    else:
        alpha = int(sca["marker alpha"] * 255)
        colork = pg.mkColor("#000000")
        colork.setAlpha(alpha)
        brush = pg.mkBrush(colork)

    # convert to log-scale if applicable
    if gen["scale x"] == "log":
        x = np.log10(x)
    if gen["scale y"] == "log":
        y = np.log10(y)

    scatter.setData(x=x, y=y, brush=brush)
    scatter.setZValue(-50)
    return [scatter]


def get_axes_labels(plot_state, slot_states):
    gen = plot_state["general"]
    labelx = dclab.dfn.get_feature_label(gen["axis x"])
    labely = dclab.dfn.get_feature_label(gen["axis y"])
    # replace FL-? with user-defined names
    fl_names = slot_states[0]["fl names"]
    if labelx.count("FL"):
        for key in fl_names:
            if key in labelx:
                labelx = labelx.replace(key, fl_names[key])
                break
    if labely.count("FL"):
        for key in fl_names:
            if key in labely:
                labely = labely.replace(key, fl_names[key])
                break
    return labelx, labely


def set_viewbox(plot, range_x, range_y, scale_x="linear", scale_y="linear",
                padding=0):
    # Set Log scale
    plot.setLogMode(x=scale_x == "log",
                    y=scale_y == "log")
    range_x = np.array(range_x)
    range_y = np.array(range_y)
    if scale_x == "log":
        if range_x[0] <= 0:
            if range_x[1] > 10:
                range_x[0] = 1e-1
            else:
                range_x[0] = 1e-3
        range_x = np.log10(range_x)
    if scale_y == "log":
        if range_y[0] <= 0:
            if range_y[1] > 10:
                range_y[0] = 1e-1
            else:
                range_y[0] = 1e-3
        range_y = np.log10(range_y)
    # Set Range
    plot.setRange(xRange=range_x,
                  yRange=range_y,
                  padding=padding,
                  )


linestyles = {
    "solid": QtCore.Qt.SolidLine,
    "dashed": QtCore.Qt.DashLine,
    "dotted": QtCore.Qt.DotLine,
}



import codecs
import numbers
import pathlib
import pkg_resources
import time

import dclab
from PyQt5 import uic, QtCore, QtWidgets

from ..widgets import show_wait_cursor
from ..._version import version

STAT_METHODS = sorted(dclab.statistics.Statistics.available_methods.keys())
STAT_METHODS.remove("%-gated")  # This does not make sense with Pipeline


class ComputeStatistics(QtWidgets.QDialog):
    def __init__(self, parent, pipeline, *args, **kwargs):
        QtWidgets.QWidget.__init__(self, parent, *args, **kwargs)
        path_ui = pkg_resources.resource_filename(
            "shapeout2.gui.compute", "comp_stats.ui")
        uic.loadUi(path_ui, self)
        # for external statistics
        self.path = None
        # set pipeline
        self.pipeline = pipeline
        # Signals
        self.pushButton_path.clicked.connect(self.on_browse)
        self.comboBox.currentIndexChanged.connect(self.on_combobox)
        # Populate statistics methods
        self.listWidget_stats.clear()
        for meth in STAT_METHODS:
            wid = QtWidgets.QListWidgetItem(meth)
            wid.setCheckState(2)
            self.listWidget_stats.addItem(wid)
        # Populate filter ray comboBox
        self.comboBox_filter_ray.clear()
        self.comboBox_filter_ray.addItem("No filtering", None)
        for ii, slot in enumerate(pipeline.slots):
            if slot.slot_used:
                raytext = "Ray {} ({})".format(ii, slot.name)
                self.comboBox_filter_ray.addItem(raytext, slot.identifier)
        # initialize rest
        if len(self.pipeline.slots) == 0:
            self.comboBox.setCurrentIndex(1)
        else:
            self.comboBox.setCurrentIndex(0)
        self.on_combobox()  # computes self.features

    def done(self, r):
        if r:
            success = self.export_statistics()
        else:
            success = True
        if success:
            super(ComputeStatistics, self).done(r)

    @show_wait_cursor
    @QtCore.pyqtSlot()
    def export_statistics(self):
        """Export statistics to .tsv"""
        # get features
        features = []
        for ii in range(self.listWidget_features.count()):
            if self.listWidget_features.item(ii).checkState() == 2:
                features.append(self.features[ii])
        # get methods
        methods = []
        for ii in range(self.listWidget_stats.count()):
            if self.listWidget_stats.item(ii).checkState() == 2:
                methods.append(STAT_METHODS[ii])

        prog = QtWidgets.QProgressDialog("Computing statistics...", "Abort", 1,
                                         1, self)
        prog.setMinimumDuration(0)
        time.sleep(0.01)
        prog.setValue(0)
        # compute statistics
        values = []
        if self.comboBox.currentIndex() == 0:
            # from pipeline
            prog.setMaximum(len(self.pipeline.slots))
            for slot_index in range(len(self.pipeline.slots)):
                slot = self.pipeline.slots[slot_index]
                if slot.slot_used:  # only export slots "used" (#15)
                    ds = self.pipeline.get_dataset(slot_index)
                    h, v = dclab.statistics.get_statistics(ds,
                                                           methods=methods,
                                                           features=features)
                    h = ["Path", "Slot", "Name"] + h
                    v = ["{}".format(ds.path), slot_index, slot.name] + v
                    values.append(v)
                if prog.wasCanceled():
                    break
                prog.setValue(slot_index + 1)
                QtWidgets.QApplication.processEvents()
        else:
            # from path
            inpath = pathlib.Path(self.path)
            files = sorted(inpath.rglob("*.rtdc"))
            prog.setMaximum(len(files))
            for ii, pp in enumerate(files):
                ds = dclab.new_dataset(pp)
                title = ds.title
                slot_id = self.comboBox_filter_ray.currentData()
                if slot_id is not None:
                    ds = self.pipeline.apply_filter_ray(rtdc_ds=ds,
                                                        slot_id=slot_id)
                    title += " ({})".format(slot_id)
                h, v = dclab.statistics.get_statistics(ds,
                                                       methods=methods,
                                                       features=features)
                h = ["Path", "Name"] + h
                v = ["{}".format(pp), title] + v
                values.append(v)
                prog.setValue(ii + 1)
                QtWidgets.QApplication.processEvents()
        # Output path
        opath, _ = QtWidgets.QFileDialog.getSaveFileName(
            self, 'Save statistics', '', 'tab-separated values (*.tsv)')
        if not opath:
            # Abort export
            return False
        elif not opath.endswith(".tsv"):
            opath += ".tsv"
        opath = pathlib.Path(opath)
        # Header
        header = ["Statistics Output",
                  "Shape-Out {}".format(version),
                  "",
                  "\t".join(h),
                  ]
        # Data
        data = []
        for v in values:
            line = []
            for vi in v:
                if (isinstance(vi, numbers.Real)
                        and not isinstance(vi, numbers.Integral)):
                    line.append("{:.5e}".format(vi))
                else:
                    line.append("{}".format(vi))
            data.append("\t".join(line))
        # Write BOM
        with opath.open("wb") as fd:
            fd.write(codecs.BOM_UTF8)
        # Write rest
        with opath.open("a", encoding="utf-8") as fd:
            for line in header:
                fd.write("# " + line + "\r\n")
            for line in data:
                fd.write(line + "\r\n")
        return True  # True means success

    def on_browse(self):
        out = QtWidgets.QFileDialog.getExistingDirectory(self,
                                                         'Export directory')
        if out:
            self.path = out
            self.lineEdit_path.setText(self.path)
            self.comboBox.setCurrentIndex(1)
        else:
            self.path = None
            self.comboBox.setCurrentIndex(0)

    def on_combobox(self):
        if self.comboBox.currentIndex() == 1:
            # Datasets from a folder
            self.widget_other.show()
            if self.path is None:
                self.on_browse()
            if self.path:
                self.update_feature_list(use_pipeline=False)
                # else, on_combobox is triggered again
        else:
            # Datasets from current session
            self.widget_other.hide()
            self.update_feature_list(use_pipeline=True)

    def update_feature_list(self, use_pipeline=True):
        if use_pipeline:
            self.features = self.pipeline.get_features(scalar=True,
                                                       union=True,
                                                       label_sort=True)
        else:
            # We want to compute statistics from a folder
            # TODO:
            # - Add ml_score_??? features

            # label-sorted features
            features = dclab.dfn.scalar_feature_names
            labs = [dclab.dfn.get_feature_label(f) for f in features]
            lf = sorted(zip(labs, features))
            self.features = [it[1] for it in lf]

        self.listWidget_features.clear()
        for feat in self.features:
            wid = QtWidgets.QListWidgetItem(dclab.dfn.get_feature_label(feat))
            wid.setCheckState(0)
            self.listWidget_features.addItem(wid)



"""Test of data set functionalities"""
import pathlib

from PyQt5 import QtCore

from shapeout2.gui.main import ShapeOut2
from shapeout2 import session
import pytest


@pytest.fixture(autouse=True)
def run_around_tests():
    # Code that will run before your test, for example:
    session.clear_session()
    # A test function will be run at this point
    yield
    # Code that will run after your test, for example:
    session.clear_session()


def test_simple(qtbot):
    """Open the main window and close it again"""
    main_window = ShapeOut2()
    main_window.close()


def test_matrix_slots(qtbot):
    mw = ShapeOut2()
    qtbot.addWidget(mw)

    # add a dataslot
    path = pathlib.Path(__file__).parent / "data" / "calibration_beads_47.rtdc"
    mw.add_dataslot(paths=[path])
    # add another one
    mw.add_dataslot(paths=[path])

    assert len(mw.pipeline.slot_ids) == 2, "we added those"
    assert len(mw.pipeline.filter_ids) == 1, "automatically added"

    # activate a dataslot
    slot_id = mw.pipeline.slot_ids[0]
    filt_id = mw.pipeline.filter_ids[0]
    em = mw.block_matrix.get_widget(slot_id, filt_id)
    qtbot.mouseClick(em, QtCore.Qt.LeftButton)
    # did that work?
    assert mw.pipeline.is_element_active(slot_id, filt_id)
    slot_id2 = mw.pipeline.slot_ids[1]
    assert not mw.pipeline.is_element_active(slot_id2, filt_id)

    # remove a dataslot
    wd = mw.block_matrix.get_widget(slot_id=slot_id)
    wd.action_remove()
    assert not mw.pipeline.is_element_active(slot_id2, filt_id)



import pkg_resources

from PyQt5 import uic, QtWidgets, QtCore


class MatrixElement(QtWidgets.QWidget):
    _quick_view_instance = None
    quickview_selected = QtCore.pyqtSignal()
    element_changed = QtCore.pyqtSignal()

    def __init__(self):
        QtWidgets.QWidget.__init__(self)
        path_ui = pkg_resources.resource_filename(
            "shapeout2.gui.matrix", "dm_element.ui")
        uic.loadUi(path_ui, self)

        self.active = False
        self.enabled = True
        self.invalid = False

        self.update_content()

    def __getstate__(self):
        state = {"active": self.active and not self.invalid,
                 "enabled": self.enabled,
                 "invalid": self.invalid}
        return state

    def __setstate__(self, state):
        self.active = state["active"] and not state["invalid"]
        self.enabled = state["enabled"]
        self.invalid = state["invalid"]
        self.update_content()

    def has_quickview(self):
        curinst = MatrixElement._quick_view_instance
        if curinst is self:
            quickview = True
        else:
            quickview = False
        return quickview

    def mousePressEvent(self, event):
        # toggle selection
        if not self.invalid:
            if event.modifiers() == QtCore.Qt.ShiftModifier:
                quickview = not self.has_quickview()
            else:
                self.active = not self.active
                quickview = False
                self.element_changed.emit()
            self.update_content(quickview)
            event.accept()

    def set_active(self, b=True):
        state = self.__getstate__()
        state["active"] = b
        self.__setstate__(state)

    def update_content(self, quickview=False):
        if self.invalid:
            color = "#DCDCDC"  # gray
            label = "invalid"
            tooltip = "Incompatible filter settings"
        elif self.active and self.enabled:
            color = "#86E789"  # green
            label = "active"
            tooltip = "Click to deactivate"
        elif self.active and not self.enabled:
            color = "#C9DAC9"  # gray-green
            label = "active\n(unused)"
            tooltip = "Click to deactivate"
        elif not self.active and self.enabled:
            color = "#EFEFEF"  # light gray
            label = "inactive"
            tooltip = "Click to activate"
        else:
            color = "#DCDCDC"  # gray
            label = "inactive\n(unused)"
            tooltip = "Click to activate"

        if not self.invalid:
            if self.has_quickview():
                do_quickview = True
            elif quickview:
                curinst = MatrixElement._quick_view_instance
                # reset color of old quick view instance
                if curinst is not None and self is not curinst:
                    MatrixElement._quick_view_instance = None
                    try:
                        curinst.update_content()
                    except RuntimeError:
                        # element has been deleted
                        pass
                MatrixElement._quick_view_instance = self
                do_quickview = True
            else:
                do_quickview = False
            if do_quickview:
                color = "#F0A1D6"
                label += "\n(QV)"
                self.quickview_selected.emit()
            else:
                tooltip += "\nShift+Click for Quick View"

        self.setStyleSheet("background-color:{}".format(color))
        self.label.setStyleSheet("background-color:{}".format(color))
        self.label.setText(label)
        self.setToolTip(tooltip)
        self.label.setToolTip(tooltip)



"""Test of filter functionalities"""
import copy
import pathlib
import tempfile

import dclab
import numpy as np
from PyQt5 import QtCore

from shapeout2.gui.main import ShapeOut2
from shapeout2 import session
import pytest


@pytest.fixture(autouse=True)
def run_around_tests():
    # Code that will run before your test, for example:
    session.clear_session()
    # A test function will be run at this point
    yield
    # Code that will run after your test, for example:
    session.clear_session()


def test_filter_min_max_inf(qtbot):
    # generate fake dataset
    path = pathlib.Path(__file__).parent / "data" / "calibration_beads_47.rtdc"
    with dclab.new_dataset(path) as ds:
        config = copy.deepcopy(ds.config)

    tmp = tempfile.mktemp(".rtdc", prefix="example_filter_inf_")
    ds2 = dclab.new_dataset({"deform": np.linspace(0, .01, 100),
                             "area_um": np.linspace(20, 200, 100),
                             "area_ratio": np.linspace(1, 1.1, 100)
                             })
    ds2.config.update(config)
    ds2["area_ratio"][0] = np.inf
    ds2["area_ratio"][1] = np.nan
    ds2.export.hdf5(tmp, features=["area_um", "deform", "area_ratio"])

    mw = ShapeOut2()
    qtbot.addWidget(mw)

    # add the file
    mw.add_dataslot(paths=[tmp])

    assert len(mw.pipeline.slot_ids) == 1, "we added that"
    assert len(mw.pipeline.filter_ids) == 1, "automatically added"

    # open the filter edit in the Analysis View
    fe = mw.block_matrix.get_widget(filt_plot_id=mw.pipeline.filter_ids[0])
    qtbot.mouseClick(fe.toolButton_modify, QtCore.Qt.LeftButton)

    # box filtering
    wf = mw.widget_ana_view.widget_filter
    # enable selection
    qtbot.mouseClick(wf.toolButton_moreless, QtCore.Qt.LeftButton)
    # find the porosity item and click the checkbox
    rc = wf._box_range_controls["area_ratio"]
    qtbot.mouseClick(rc.checkBox, QtCore.Qt.LeftButton)
    # disable selection
    qtbot.mouseClick(wf.toolButton_moreless, QtCore.Qt.LeftButton)

    # check that the range control does not have all-zero values
    rcstate = rc.__getstate__()
    assert rcstate["start"] != 0
    assert rcstate["end"] != 0
    # only approximate (b/c they were converted on the range scale)
    assert np.allclose(rcstate["start"], ds2["area_ratio"][2], rtol=1e-4)
    assert np.allclose(rcstate["end"], 1.1, rtol=1e-4)



import pkg_resources

from PyQt5 import uic, QtCore, QtWidgets


class Preferences(QtWidgets.QDialog):
    """Implements the plotting pipeline using pyqtgraph"""
    instances = {}

    def __init__(self, parent, *args, **kwargs):
        QtWidgets.QWidget.__init__(self, parent=parent, *args, **kwargs)
        path_ui = pkg_resources.resource_filename(
            "shapeout2.gui", "preferences.ui")
        uic.loadUi(path_ui, self)
        self.settings = QtCore.QSettings()
        self.parent = parent

        #: configuration keys, corresponding widgets, and defaults
        self.config_pairs = [
            ["advanced/developer mode", self.advanced_developer_mode, 0],
            ["advanced/check pyqtgraph version",
             self.advanced_check_pyqtgraph_version, 1],
            ["dcor/api key", self.dcor_api_key, ""],
            ["dcor/servers", self.dcor_servers, ["dcor.mpl.mpg.de"]],
            ["dcor/use ssl", self.dcor_use_ssl, 1],
            ["check for updates", self.general_check_for_updates, 1],
        ]
        self.reload()

        # signals
        btn_apply = self.buttonBox.button(QtWidgets.QDialogButtonBox.Apply)
        btn_apply.clicked.connect(self.on_apply)
        btn_ok = self.buttonBox.button(QtWidgets.QDialogButtonBox.Ok)
        btn_ok.clicked.connect(self.on_apply)
        btn_restore = self.buttonBox.button(
            QtWidgets.QDialogButtonBox.RestoreDefaults)
        btn_restore.clicked.connect(self.on_restore)

    def reload(self):
        """Read configuration or set default parameters"""
        for key, widget, default in self.config_pairs:
            value = self.settings.value(key, default)
            if isinstance(widget, QtWidgets.QCheckBox):
                widget.setChecked(bool(int(value)))
            elif isinstance(widget, QtWidgets.QLineEdit):
                widget.setText(value)
            elif widget is self.dcor_servers:
                self.dcor_servers.clear()
                self.dcor_servers.addItems(value)
                self.dcor_servers.setCurrentIndex(0)
            else:
                raise NotImplementedError("No rule for '{}'".format(key))

        # peculiarities of developer mode
        devmode = bool(int(self.settings.value("advanced/developer mode", 0)))
        self.dcor_use_ssl.setVisible(devmode)  # show "use ssl" in dev mode

    @QtCore.pyqtSlot()
    def on_apply(self):
        """Save current changes made in UI to settings and reload UI"""
        for key, widget, default in self.config_pairs:
            if isinstance(widget, QtWidgets.QCheckBox):
                value = int(widget.isChecked())
            elif isinstance(widget, QtWidgets.QLineEdit):
                value = widget.text().strip()
            elif widget is self.dcor_servers:
                curtext = self.dcor_servers.currentText()
                items = self.settings.value(key, default)
                if curtext in items:
                    items.remove(curtext)
                items.insert(0, curtext)
                value = items
            else:
                raise NotImplementedError("No rule for '{}'".format(key))
            self.settings.setValue(key, value)

        # reload UI to give visual feedback
        self.reload()

    @QtCore.pyqtSlot()
    def on_restore(self):
        self.settings.clear()
        self.reload()

    @QtCore.pyqtSlot(bool)
    def on_boolean(self, b):
        widget = self.sender()
        key = self.get_key_from_widget(widget)
        self.settings.setValue(key, int(b))
        if widget == self.advanced_developer_mode:
            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Information)
            msg.setText("Please restart Shape-Out for the changes to take "
                        + "effect.")
            msg.setWindowTitle("Restart Shape-Out")
            msg.exec_()



import pkg_resources

import numpy as np
from PyQt5 import uic, QtCore, QtWidgets


#: Precision for these features (``data``) should not go below this value.
#: (Precision is set automatically based on data range)
SPIN_CONTROL_PRECISION = {
    "area_ratio": 3,
    "aspect": 3,
    "bright_avg": 1,
    "bright_sd": 3,
}


class RangeControl(QtWidgets.QWidget):
    #: Emitted when the range changed
    range_changed = QtCore.pyqtSignal(float, float)

    def __init__(self, parent, label="feature", checkbox=True, integer=False,
                 data=None, *args, **kwargs):
        QtWidgets.QWidget.__init__(self, parent, *args, **kwargs)
        path_ui = pkg_resources.resource_filename(
            "shapeout2.gui.widgets", "rangecontrol.ui")
        uic.loadUi(path_ui, self)

        # arbitrary data
        self.data = data

        # default values for limits
        self.minimum = -100
        self.maximum = +100

        # label on top of control
        self.setLabel(label)

        # enable/disable checkbox
        if not checkbox:
            self.checkBox.hide()

        # integer-valued control
        if integer:
            self.doubleSpinBox_min.setDecimals(0)
            self.doubleSpinBox_max.setDecimals(0)
        self.is_integer = integer

        # reduce font size of name
        font = self.label.font()
        font.setPointSize(font.pointSize()-1)
        self.label.setFont(font)

        # signals
        self.range_slider.rangeChanged.connect(self.on_range)
        self.doubleSpinBox_min.valueChanged.connect(self.on_spinbox)
        self.doubleSpinBox_max.valueChanged.connect(self.on_spinbox)

        # call show to make sure slider is updated
        self.show()

    def __getstate__(self):
        state = {
            "active": self.checkBox.isChecked(),
            "start": self.doubleSpinBox_min.value(),
            "end": self.doubleSpinBox_max.value(),
        }
        return state

    def __setstate__(self, state):
        self.checkBox.setChecked(state["active"])
        self.doubleSpinBox_min.blockSignals(True)
        self.doubleSpinBox_max.blockSignals(True)
        self.doubleSpinBox_min.setValue(state["start"])
        self.doubleSpinBox_max.setValue(state["end"])
        self.doubleSpinBox_min.blockSignals(False)
        self.doubleSpinBox_max.blockSignals(False)
        self.map_spin_values_to_range_slider()

    @QtCore.pyqtSlot(float, float)
    def map_spin_values_to_range_slider(self, vmin=None, vmax=None):
        # limits
        lmin = self.minimum
        lmax = self.maximum
        # spin values
        if vmin is None:
            vmin = self.doubleSpinBox_min.value()
        if vmax is None:
            vmax = self.doubleSpinBox_max.value()
        # range slider limits
        rmin = self.range_slider.min()
        rmax = self.range_slider.max()
        # compute values
        dr = rmax - rmin
        dl = lmax - lmin
        if dl * dr == 0:
            hmin = hmax = 0
        else:
            hmin = rmin + (vmin - lmin) / dl * dr
            hmax = rmax - (lmax - vmax) / dl * dr
        if hmin < rmin:
            hmin = 0
        if hmax > rmax:
            hmax = self.range_slider._INT_NUM
        self.range_slider.blockSignals(True)
        self.range_slider.setRange(hmin, hmax)
        self.range_slider.blockSignals(False)
        # make range selection stick tight to edges
        if hmin < 10:
            hmin = 0
        if hmin > self.range_slider._INT_NUM - 10:
            hmax = self.range_slider._INT_NUM
        return hmin, hmax

    @QtCore.pyqtSlot(int, int)
    def map_range_slider_to_spin_values(self, hmin=None, hmax=None):
        """Return the respective value of the current range

        Range limits are defined by
        - self.doubleSpinBox_min.minimum()
        - self.doubleSpinBox_min.maximum()
        """
        # limits
        lmin = self.minimum
        lmax = self.maximum
        # range slider limits
        rmin = self.range_slider.min()
        rmax = self.range_slider.max()
        # range slider handles
        if hmin is None:
            hmin = self.range_slider.start()
        if hmax is None:
            hmax = self.range_slider.end()
        # compute values
        dr = rmax - rmin
        dl = lmax - lmin
        vmin = lmin + (hmin - rmin) / dr * dl
        vmax = lmax - (rmax - hmax) / dr * dl

        self.doubleSpinBox_min.blockSignals(True)
        self.doubleSpinBox_max.blockSignals(True)
        self.doubleSpinBox_min.setValue(vmin)
        self.doubleSpinBox_max.setValue(vmax)
        self.doubleSpinBox_min.blockSignals(False)
        self.doubleSpinBox_max.blockSignals(False)
        return vmin, vmax

    def on_range(self):
        self.map_range_slider_to_spin_values()
        self.range_changed.emit(self.doubleSpinBox_min.value(),
                                self.doubleSpinBox_max.value())

    def on_spinbox(self):
        self.map_spin_values_to_range_slider()
        self.range_changed.emit(self.doubleSpinBox_min.value(),
                                self.doubleSpinBox_max.value())

    def reset_range(self):
        self.doubleSpinBox_min.setValue(self.minimum)
        self.doubleSpinBox_max.setValue(self.maximum)

    def setCheckable(self, b=True):
        self.checkBox.setVisible(b)

    def setInteger(self, b=True):
        self.is_integer = b

    def setLabel(self, label):
        if label:
            self.label.setText(label)
            self.label.show()
        else:
            self.label.hide()

    def setLimits(self, vmin, vmax, hard_limit=False):
        """Set the limits of the range control

        Parameters
        ----------
        vmin, vmax: float
            Minimum and maximum values
        hard_limit: bool
            If set to True, the spin controls will have a hard limit
            that matches vmin and vmax. If False (default), the
            limit of the spin controls is larger, giving the user
            a broader range.
        """
        self.minimum = vmin
        self.maximum = vmax

        if hard_limit:
            vminh = vmin
            vmaxh = vmax
        else:
            # estimate based on number of digits
            if vmin >= 0:
                vminh = 0
            else:
                lmin = np.int(np.ceil(np.log10(np.abs(vmin)))) + 1
                vminh = -10**lmin
            if vmax > 0:
                lmax = np.int(np.ceil(np.log10(vmax))) + 1
                vmaxh = 10**lmax
            else:
                vmaxh = 0

        self.setSpinLimits(vmin=vminh, vmax=vmaxh)
        # slider values
        self.map_spin_values_to_range_slider()

    def setSpinLimits(self, vmin, vmax):
        """Sets spin control limits and precision

        Notes
        -----
        The precision is set automatically from the min/max
        peak-to-peak size. If this precision is not high
        enough for a specific application, you can set the
        ``data`` attribute upon initialization and add the
        precision (in decimals) to :const:`SPIN_CONTROL_PRECISION`
        via ``SPIN_CONTROL_PRECISION[data] = precision``.
        """
        # min/max
        self.doubleSpinBox_min.setMinimum(vmin)
        self.doubleSpinBox_max.setMinimum(vmin)
        self.doubleSpinBox_min.setMaximum(vmax)
        self.doubleSpinBox_max.setMaximum(vmax)

        # decimals
        if not self.is_integer:
            if vmax == vmin:
                dec = 1
            else:
                # two significant digits
                dec = int(np.ceil(np.log10(1/np.abs(vmax-vmin)))) + 5
                if dec <= 0:
                    dec = 1
            if self.data in SPIN_CONTROL_PRECISION:
                dec = max(dec, SPIN_CONTROL_PRECISION[self.data])
            self.doubleSpinBox_min.setDecimals(dec)
            self.doubleSpinBox_max.setDecimals(dec)
            self.doubleSpinBox_min.setSingleStep(10**-dec)
            self.doubleSpinBox_max.setSingleStep(10**-dec)



import copy

import dclab
import numpy as np

from ..util import hashobj


DEFAULT_STATE = {
    "identifier": "no default",
    "layout": {
        "column count": 2,
        "division": "multiscatter+contour",
        "label plots": True,
        "name": "no default",  # overridden by __init__
        "size x": 400,
        "size y": 400,
    },
    "general": {
        "auto range": True,  # this overrides range x and range y
        "axis x": "area_um",
        "axis y": "deform",
        "isoelastics": True,  # display isoelasticity lines
        "kde": "histogram",  # see dclab.kde_methods.methods
        "range x": [0, 0],  # equal means no preference
        "range y": [0, 0],
        "scale x": "linear",
        "scale y": "linear",
    },
    "scatter": {
        "colormap": "viridis",  # only applies when hue is "kde" or "feature"
        "downsample": True,
        "downsampling value": 5000,
        "enabled": True,
        "hue feature": "bright_avg",  # which feature to use, if set
        "hue max": 1,
        "hue min": 0,
        "marker alpha": 0.3,  # alpha value for feature-hue plots
        "marker hue": "kde",  # hue defined by: kde, dataset, feature, none
        "marker size": 3.0,  # marker size [pt]
        "show event count": True,  # display event count
    },
    "contour": {
        "enabled": True,
        "legend": False,  # display plot legend
        "line widths": [3.0, 1.5],  # contour line widths [pt]
        "line styles": ["solid", "dashed"],
        "percentiles": [95.0, 50.0],
        "spacing x": 2,  # spacing for "axis x" and linear "scale x"
        "spacing y": 0.005,  # spacing for "axis y" and linear "scale y"
    }
}

_kde_methods = sorted(dclab.kde_methods.methods.keys())
_kde_methods.remove("none")  # does not make sense here

STATE_OPTIONS = {
    "identifier": str,
    "layout": {
        "column count": int,
        "division": ["each", "merge", "multiscatter+contour"],
        "label plots": bool,
        "name": str,
        "size x": float,
        "size y": float,
    },
    "general": {
        "auto range": bool,
        # This is not entirely correct, because `ml_score_???`
        # should also work.
        "axis x": dclab.dfn.scalar_feature_names,
        "axis y": dclab.dfn.scalar_feature_names,
        "isoelastics": bool,
        "kde": _kde_methods,
        "range x": (float,),
        "range y": (float,),
        "scale x": ["linear", "log"],
        "scale y": ["linear", "log"],
    },
    "scatter": {
        "colormap": ["bipolar", "grayblue", "graygreen", "grayorange",
                     "grayred", "inferno", "plasma", "viridis"],
        "downsampling": bool,
        "downsampling value": int,
        "enabled":  bool,
        # This is not entirely correct, because `ml_score_???`
        # should also work.
        "hue feature": dclab.dfn.scalar_feature_names,
        "hue max": float,
        "hue min": float,
        "marker alpha": float,
        "marker hue": ["dataset", "kde", "feature", "none"],
        "marker size": float,
        "show event count": bool,
    },
    "contour": {
        "enabled":  bool,
        "legend": bool,
        "line widths": (float,),
        "line styles": (["solid", "dashed", "dotted"],),
        "percentiles": (float,),
        "spacing x": float,
        "spacing y": float,
    }
}


class Plot(object):
    """Handles plotting information in a pipeline"""
    _instance_counter = 0
    _instances = {}

    def __init__(self, identifier=None, name=None):
        Plot._instance_counter += 1
        if identifier is None:
            identifier = "Plot_{}".format(Plot._instance_counter)
            while identifier in Plot._instances:
                Plot._instance_counter += 1
                identifier = "Plot_{}".format(Plot._instance_counter)

        # initially, set default state
        self._state = copy.deepcopy(DEFAULT_STATE)

        if name is None:
            name = identifier
        #: unique identifier of the plot
        self.identifier = identifier
        #: user-defined name of the plot
        self.name = name
        if identifier in Plot._instances:
            raise ValueError("Plot with identifier "
                             + "'{}' already exists!".format(identifier))
        Plot._instances[identifier] = self

    def __getstate__(self):
        state = copy.deepcopy(self._state)
        state["identifier"] = self.identifier
        return state

    def __repr__(self):
        repre = "<Pipeline Plot '{}' at {}>".format(self.identifier,
                                                    hex(id(self)))
        return repre

    def __setstate__(self, state):
        state = copy.deepcopy(state)
        if self.identifier != state["identifier"]:
            raise ValueError("Identifier mismatch: '{}' vs. '{}'".format(
                self.identifier, state["identifier"]))
        # disallow nan-values
        if np.any(np.isinf(state["general"]["range x"])):
            state["general"]["range x"] = [0, 0]
        if np.any(np.isinf(state["general"]["range y"])):
            state["general"]["range y"] = [0, 0]
        self._state = state

    @staticmethod
    def get_instances():
        return Plot._instances

    @staticmethod
    def get_plot(identifier):
        """Get the plot with the given identifier.

        Notes
        -----
        Creates the plot if it does not exist.
        """
        if identifier in Plot._instances:
            f = Plot._instances[identifier]
        else:
            f = Plot(identifier=identifier)
        return f

    @property
    def hash(self):
        """Return the hash of the plot"""
        return hashobj(self.__getstate__())

    @property
    def name(self):
        return self._state["layout"]["name"]

    @name.setter
    def name(self, value):
        self._state["layout"]["name"] = value



"""Convenience methods to retrieve meta data from .rtdc files"""
import functools
import dclab


def get_info(path, section, key):
    config = get_rtdc_config(path)
    return config[section][key]


def get_repr(path, append_path=False):
    """representative string of a dataset"""
    exp = get_rtdc_config(path)["experiment"]
    rep = "{} #{} ({} {})".format(exp["sample"],
                                  exp["run index"],
                                  exp["date"],
                                  exp["time"])
    if append_path:
        rep += "\n{}".format(path)
    return rep


@functools.lru_cache(maxsize=100)
def get_rtdc_config(path):
    with dclab.new_dataset(path) as ds:
        config = ds.config.copy()
    return config


@functools.lru_cache(maxsize=100)
def get_rtdc_features(path, scalar=True, only_loaded=False):
    """Return available features in a dataset"""
    av_feat = []
    with dclab.new_dataset(path) as ds:
        if scalar:
            features = ds.features_scalar
        else:
            features = ds.features
        for feat in features:
            if only_loaded:
                if feat in ds.features_loaded:
                    av_feat.append(feat)
            else:
                if feat in ds:
                    av_feat.append(feat)
    return av_feat


def get_rtdc_features_bulk(paths, scalar=True):
    """Return available features for a list of dataset paths"""
    features = []
    for pp in paths:
        features += get_rtdc_features(path=pp, scalar=scalar)
    return sorted(set(features))


@functools.lru_cache(maxsize=10000)
def get_rtdc_features_minmax(path, *features):
    """Return dict with min/max of scalar features in a dataset"""
    mmdict = {}
    with dclab.new_dataset(path) as ds:
        if len(features) == 0:
            features = ds.features_loaded
        for feat in features:
            assert dclab.dfn.scalar_feature_exists(feat)
            if feat in ds:
                mmdict[feat] = ds[feat].min(), ds[feat].max()
    return mmdict


def get_rtdc_features_minmax_bulk(paths, features=[]):
    """Perform `get_rtdc_features_minmax` on a list of paths

    Parameters
    ----------
    paths: list of str or list of pathlib.Path
        Paths to measurement files
    features: list of str or empty list
        Names of the features to compute the min/max values for.
        If empty, all loaded features will be used.
    """
    mmdict = {}
    for pp in paths:
        mmdi = get_rtdc_features_minmax(pp, *features)
        for feat in mmdi:
            if feat in mmdict:
                fmin = min(mmdict[feat][0], mmdi[feat][0])
                fmax = max(mmdict[feat][1], mmdi[feat][1])
                mmdict[feat] = (fmin, fmax)
            else:
                mmdict[feat] = mmdi[feat]
    return mmdict



def main(splash=True):
    import os
    import pkg_resources
    import sys
    import time

    from PyQt5.QtWidgets import QApplication
    # import before creating application
    import pyqtgraph  # noqa: F401

    app = QApplication(sys.argv)
    imdir = pkg_resources.resource_filename("shapeout2", "img")

    if splash:
        from PyQt5.QtWidgets import QSplashScreen
        from PyQt5.QtGui import QPixmap
        splash_path = os.path.join(imdir, "splash.png")
        splash_pix = QPixmap(splash_path)
        splash = QSplashScreen(splash_pix)
        splash.setMask(splash_pix.mask())
        splash.show()
        # make sure Qt really displays the splash screen
        time.sleep(.07)
        app.processEvents()

    from PyQt5 import QtCore, QtGui
    from .gui import ShapeOut2

    # Set Application Icon
    icon_path = os.path.join(imdir, "icon.png")
    app.setWindowIcon(QtGui.QIcon(icon_path))

    # Use dots as decimal separators
    QtCore.QLocale.setDefault(QtCore.QLocale(QtCore.QLocale.C))

    window = ShapeOut2()

    if splash:
        splash.finish(window)

    sys.exit(app.exec_())


if __name__ == "__main__":
    main()



import copy
import pkg_resources

from PyQt5 import uic, QtCore, QtWidgets

import dclab

from ...pipeline import Filter

from ..widgets import RangeControl
from .. import idiom


class FilterPanel(QtWidgets.QWidget):
    #: Emitted when a shapeout2.pipeline.Filter is to be changed
    filter_changed = QtCore.pyqtSignal(dict)
    #: Emitted when the pipeline is to be changed
    pipeline_changed = QtCore.pyqtSignal(dict)
    #: Emitted when the user wants to create a new polygon filter
    request_new_polygon_filter = QtCore.pyqtSignal()

    def __init__(self, *args, **kwargs):
        QtWidgets.QWidget.__init__(self)
        path_ui = pkg_resources.resource_filename(
            "shapeout2.gui.analysis", "ana_filter.ui")
        uic.loadUi(path_ui, self)
        # current Shape-Out 2 pipeline
        self._pipeline = None
        self.setUpdatesEnabled(False)
        #: contains the range widgets for the box filters
        self._box_range_controls = {}
        self._populate_box_filters()
        self._polygon_checkboxes = {}
        self.update_polygon_filters()
        self.toolButton_duplicate.clicked.connect(self.on_duplicate_filter)
        self.toolButton_remove.clicked.connect(self.on_remove_filter)
        self.pushButton_apply.clicked.connect(self.write_filter)
        self.pushButton_reset.clicked.connect(self.update_content)
        self.comboBox_filters.currentIndexChanged.connect(self.update_content)
        self.toolButton_moreless.clicked.connect(self.on_moreless)
        self.label_box_edit.setVisible(False)
        self._box_edit_view = False
        self.update_content()
        self.setUpdatesEnabled(True)

    def __getstate__(self):
        state = {
            "filter used": self.checkBox_enable.isChecked(),
            "identifier": self.current_filter.identifier,
            "limit events bool": self.checkBox_limit.isChecked(),
            "limit events num": self.spinBox_limit.value(),
            "name": self.lineEdit_name.text(),
            "remove invalid events": self.checkBox_invalid.isChecked(),
        }
        # box filters
        box = {}
        for feat in self.active_box_features:
            rc = self._box_range_controls[feat]
            box[feat] = rc.__getstate__()
        state["box filters"] = box
        # polygon filters
        pflist = []
        for key in self._polygon_checkboxes:
            if self._polygon_checkboxes[key].isChecked():
                pflist.append(key)
        state["polygon filters"] = pflist
        return state

    def __setstate__(self, state):
        if self.current_filter.identifier != state["identifier"]:
            raise ValueError("Filter identifier mismatch!")
        self.checkBox_enable.setChecked(state["filter used"])
        self.lineEdit_name.setText(state["name"])
        self.checkBox_limit.setChecked(state["limit events bool"])
        self.spinBox_limit.setValue(state["limit events num"])
        self.checkBox_invalid.setChecked(state["remove invalid events"])
        # box filters
        box = state["box filters"]
        for feat in self._box_range_controls:
            rc = self._box_range_controls[feat]
            if feat in box:
                rc.show()
                rc.__setstate__(box[feat])
            else:
                rc.hide()
                rc.reset_range()

        # polygon filters
        pflist = state["polygon filters"]
        for key in self._polygon_checkboxes:
            if key in pflist:
                self._polygon_checkboxes[key].setChecked(True)
            else:
                self._polygon_checkboxes[key].setChecked(False)

    def _populate_box_filters(self):
        """Dynamically update available pipeline box filters

        This method can be called multiple times. If called multiple
        times, additional features that were not there before
        (e.g. `ml_score_???`) are added.
        """
        feats, labs = self.get_features_labels()

        self.verticalLayout_box.setAlignment(QtCore.Qt.AlignTop)

        for lab, feat in sorted(zip(labs, feats)):
            integer = True if feat in idiom.INTEGER_FEATURES else False
            if feat not in self._box_range_controls:
                # Create the control
                rc = RangeControl(
                    self,
                    checkbox=False,  # checkbox is used in on_moreless
                    integer=integer,
                    label=lab,
                    data=feat)
                rc.checkBox.setChecked(False)
                rc.setVisible(False)
                # Insert the control at the correct position (label-sorted)
                rcf = list(self._box_range_controls.keys())
                rcl = [dclab.dfn.get_feature_label(ft) for ft in rcf]
                index = sorted(rcl + [lab]).index(lab) + 1  # +1 b/c new list
                self.verticalLayout_box.insertWidget(index, rc)
                self._box_range_controls[feat] = rc

    @property
    def active_box_features(self):
        """List of box-filtered features that are active"""
        act = []
        for feat, item in self._box_range_controls.items():
            if item.__getstate__()["active"]:
                act.append(feat)
        return act

    @property
    def current_filter(self):
        if self.filter_ids:
            filt_index = self.comboBox_filters.currentIndex()
            filt_id = self.filter_ids[filt_index]
            filt = Filter.get_instances()[filt_id]
        else:
            filt = None
        return filt

    @property
    def filter_ids(self):
        """List of filter identifiers"""
        if self.pipeline is not None:
            ids = [filt.identifier for filt in self.pipeline.filters]
        else:
            ids = []
        return ids

    @property
    def filter_names(self):
        """List of filter names"""
        if self.pipeline is not None:
            nms = [filt.name for filt in self.pipeline.filters]
        else:
            nms = []
        return nms

    @property
    def pipeline(self):
        return self._pipeline

    def get_features_labels(self):
        """Wrapper around pipeline with default features if empty"""
        if self.pipeline is not None and self.pipeline.num_slots != 0:
            feats, labs = self.pipeline.get_features(scalar=True,
                                                     label_sort=True,
                                                     ret_labels=True,
                                                     union=True)
        else:
            # fallback (nothing in the pipeline or no pipeline)
            features = dclab.dfn.scalar_feature_names
            labs = [dclab.dfn.get_feature_label(f) for f in features]
            lf = sorted(zip(labs, features))
            feats = [it[1] for it in lf]
            labs = [it[0] for it in lf]
        return feats, labs

    def on_duplicate_filter(self):
        # determine the new filter state
        filt_state = self.__getstate__()
        new_state = copy.deepcopy(filt_state)
        new_filt = Filter()
        new_state["identifier"] = new_filt.identifier
        new_state["name"] = new_filt.name
        new_filt.__setstate__(new_state)
        # determine the filter position
        pos = self.pipeline.filter_ids.index(filt_state["identifier"])
        self.pipeline.add_filter(new_filt, index=pos+1)
        state = self.pipeline.__getstate__()
        self.pipeline_changed.emit(state)

    def on_remove_filter(self):
        filt_state = self.__getstate__()
        self.pipeline.remove_filter(filt_state["identifier"])
        state = self.pipeline.__getstate__()
        self.pipeline_changed.emit(state)

    def on_moreless(self):
        """User wants to choose box filters"""
        if not self._box_edit_view:
            # get available features to show
            features, _ = self.get_features_labels()
            # create missing range controls if applicable (e.g. ml_score_???)
            self._populate_box_filters()
            # Show all filters shared by all datasets to the user
            for feat, rc in self._box_range_controls.items():
                if feat in features:
                    rc.setVisible(True)
                    rc.checkBox.setVisible(True)
                    rc.doubleSpinBox_min.setEnabled(False)
                    rc.doubleSpinBox_max.setEnabled(False)
                    rc.range_slider.setEnabled(False)
            self.toolButton_moreless.setText("...Finish editing")
            self.label_box_edit.setVisible(True)
            self._box_edit_view = True
        else:
            # Hide all filters that are not active
            for _, rc in self._box_range_controls.items():
                if not rc.__getstate__()["active"]:
                    rc.setVisible(False)
                rc.checkBox.setVisible(False)
                rc.doubleSpinBox_min.setEnabled(True)
                rc.doubleSpinBox_max.setEnabled(True)
                rc.range_slider.setEnabled(True)
            self.toolButton_moreless.setText("Choose filters...")
            self.label_box_edit.setVisible(False)
            self._box_edit_view = False
            # Update box filter ranges
            self.update_box_ranges()

    def set_pipeline(self, pipeline):
        self._pipeline = pipeline

    def show_filter(self, filt_id):
        self.update_content(filt_index=self.filter_ids.index(filt_id))

    def update_content(self, event=None, filt_index=None):
        if self.filter_ids:
            self.setEnabled(True)
            self.update_polygon_filters(update_state=False)
            # update combobox
            self.comboBox_filters.blockSignals(True)
            if filt_index is None:
                filt_index = self.comboBox_filters.currentIndex()
                if filt_index > len(self.filter_ids) - 1 or filt_index < 0:
                    filt_index = len(self.filter_ids) - 1
            self.comboBox_filters.clear()
            self.comboBox_filters.addItems(self.filter_names)
            self.comboBox_filters.setCurrentIndex(filt_index)
            self.comboBox_filters.blockSignals(False)
            # populate content
            filt = Filter.get_filter(identifier=self.filter_ids[filt_index])
            state = filt.__getstate__()
            self.__setstate__(state)
            self.update_box_ranges()
        else:
            self.setEnabled(False)

    def update_box_ranges(self):
        """Update the box plot filter ranges

        Parameters
        ----------
        show_features: None or list of str
            The features to show. Features in this list will become
            visible, all other features will be hidden.
            If set to `None`, no changes are performed.
        mmdict: dict
            The min/max dictionary for updating the limits.
            Each key is a feature name; each item is a tuple
            of min/max values for that feature.
        """
        if self.pipeline is not None and self.pipeline.num_slots:
            # compute min/max values
            mmdict = {}
            for feat in self.active_box_features:
                mmdict[feat] = self.pipeline.get_min_max(feat=feat)
            # update used features
            for feat in self._box_range_controls:
                rc = self._box_range_controls[feat]
                if feat in mmdict:
                    rc.setLimits(*mmdict[feat])
                if self.current_filter is not None:
                    state = self.current_filter.__getstate__()
                    if feat not in state["box filters"]:
                        # reset range to limits
                        rc.reset_range()

    def update_polygon_filters(self, update_state=True):
        """Update the layout containing the polygon filters"""
        self.verticalLayout_poly.setAlignment(QtCore.Qt.AlignTop)
        # clear layout
        for ii in reversed(range(self.verticalLayout_poly.count())):
            item = self.verticalLayout_poly.itemAt(ii).widget()
            if item is not None:
                item.hide()
                item.deleteLater()
        self._polygon_checkboxes = {}  # must come after getting the state
        if dclab.PolygonFilter.instances:
            for pf in dclab.PolygonFilter.instances:
                widget = QtWidgets.QWidget()
                hbox = QtWidgets.QHBoxLayout()
                hbox.setAlignment(QtCore.Qt.AlignLeft)
                hbox.setContentsMargins(0, 0, 0, 0)
                chb = QtWidgets.QCheckBox()
                hbox.addWidget(chb)
                hbox.addWidget(QtWidgets.QLabel(pf.name))
                widget.setLayout(hbox)
                self.verticalLayout_poly.addWidget(widget)
                self._polygon_checkboxes[pf.unique_id] = chb
        else:
            label = QtWidgets.QLabel("No polygon filters have been created "
                                     + "yet.")
            button = QtWidgets.QPushButton("Create polygon filter")
            button.clicked.connect(self.request_new_polygon_filter)
            self.verticalLayout_poly.addWidget(label)
            self.verticalLayout_poly.addWidget(button)
        # update current filters
        if update_state and self.current_filter is not None:
            self.__setstate__(self.current_filter.__getstate__())

    def write_filter(self):
        """Update the shapeout2.pipeline.Filter instance"""
        # get current index
        filter_state = self.__getstate__()
        self.filter_changed.emit(filter_state)
        self.update_content()  # update filter selection combobox



import pkg_resources

from dclab.features.emodulus.viscosity import KNOWN_MEDIA

from PyQt5 import uic, QtCore, QtWidgets

from shapeout2.gui.analysis.ana_slot import SlotPanel
from shapeout2.gui.widgets import show_wait_cursor


class BulkActionEmodulus(QtWidgets.QDialog):
    #: Emitted when the pipeline is to be changed
    pipeline_changed = QtCore.pyqtSignal(dict)

    def __init__(self, parent, pipeline, *args, **kwargs):
        QtWidgets.QWidget.__init__(self, parent, *args, **kwargs)
        path_ui = pkg_resources.resource_filename(
            "shapeout2.gui.bulk", "bulk_emodulus.ui")
        uic.loadUi(path_ui, self)
        # main
        self.parent = self.parent

        # set pipeline
        self.pipeline = pipeline

        # ui choices
        self.comboBox_medium.clear()
        choices = KNOWN_MEDIA + ["other"]
        for choice in choices:
            if choice == "CellCarrierB":
                name = "CellCarrier B"  # [sic]
            else:
                name = choice
            self.comboBox_medium.addItem(name, choice)
        self.comboBox_medium.addItem("not defined", "undefined")
        self.comboBox_medium.addItem("unchanged", "unchanged")
        self.comboBox_medium.currentIndexChanged.connect(self.on_cb_medium)
        self.comboBox_medium.setCurrentIndex(self.comboBox_medium.count()-1)

        self.comboBox_temp.clear()
        self.comboBox_temp.addItem("From feature", "feature")
        self.comboBox_temp.addItem("From meta data", "config")
        self.comboBox_temp.addItem("Manual", "manual")
        self.comboBox_temp.currentIndexChanged.connect(self.on_cb_temp)
        self.comboBox_temp.setCurrentIndex(self.comboBox_temp.count()-1)

        # buttons
        btn_ok = self.buttonBox.button(QtWidgets.QDialogButtonBox.Ok)
        btn_ok.clicked.connect(self.on_ok)

    @QtCore.pyqtSlot()
    def on_ok(self):
        self.set_emodulus_properties()
        self.update_ui()

    def on_cb_medium(self):
        """User changed medium"""
        medium = self.comboBox_medium.currentData()
        if medium in KNOWN_MEDIA + ["unchanged"]:
            self.doubleSpinBox_visc.setEnabled(False)
            self.comboBox_temp.setEnabled(True)
        else:
            self.doubleSpinBox_visc.setEnabled(True)
            self.comboBox_temp.setEnabled(False)
        self.on_cb_temp()

    def on_cb_temp(self):
        """User changed temperature"""
        temp = self.comboBox_temp.currentData()

        if not self.comboBox_temp.isEnabled() or temp in ["feature", "config"]:
            self.doubleSpinBox_temp.setEnabled(False)
        else:
            self.doubleSpinBox_temp.setEnabled(True)

    @show_wait_cursor
    @QtCore.pyqtSlot()
    def set_emodulus_properties(self):
        """Set the given emodulus properties for all datasets"""
        medium = self.comboBox_medium.currentData()
        if self.comboBox_temp.isEnabled():
            scen = self.comboBox_temp.currentData()
        else:
            scen = None
        if self.doubleSpinBox_temp.isEnabled():
            tempval = self.doubleSpinBox_temp.value()
        else:
            tempval = None
        if self.doubleSpinBox_visc.isEnabled():
            viscval = self.doubleSpinBox_visc.value()
        else:
            viscval = None

        if len(self.pipeline.slots) == 0:
            return

        for slot in self.pipeline.slots:
            ds = slot.get_dataset()

            # Use the internal sanity checks to determine whether
            # or not we can set the medium or temperature scenarios.
            valid_media = SlotPanel.get_dataset_choices_medium(ds)
            valid_scenarios = SlotPanel.get_dataset_choices_temperature(ds)
            if medium in [m[1] for m in valid_media]:
                state = slot.__getstate__()
                state["emodulus"]["emodulus medium"] = medium
                # Set the viscosity here, because unknown media are
                # available.
                if viscval is not None:
                    state["emodulus"]["emodulus viscosity"] = viscval
                slot.__setstate__(state)

            if scen in [s[1] for s in valid_scenarios]:  # scen is not None
                state = slot.__getstate__()
                state["emodulus"]["emodulus scenario"] = scen
                if tempval is not None:
                    state["emodulus"]["emodulus temperature"] = tempval
                slot.__setstate__(state)

    def update_ui(self):
        """Update all relevant parts of the main user interface"""
        state = self.pipeline.__getstate__()
        self.pipeline_changed.emit(state)



from .core import Pipeline  # noqa: F401
from .dataslot import Dataslot  # noqa: F401
from .filter import Filter  # noqa: F401
from .filter_ray import FilterRay  # noqa: F401
from .plot import Plot  # noqa: F401



import copy

import dclab
from dclab.features.emodulus.viscosity import KNOWN_MEDIA
import numpy as np

from .. import meta_tool
from ..util import hashobj


class Dataslot(object):
    """Handles datasets in a pipeline"""
    _instance_counter = 0
    _instances = {}

    def __init__(self, path, identifier=None, name=None):
        Dataslot._instance_counter += 1
        self.path = path
        self._dataset = None
        if identifier is None:
            identifier = "Dataslot_{}".format(Dataslot._instance_counter)
            while identifier in Dataslot._instances:
                Dataslot._instance_counter += 1
                identifier = "Dataslot_{}".format(Dataslot._instance_counter)
        cfg = meta_tool.get_rtdc_config(path)
        if name is None:
            name = cfg["experiment"]["sample"]
        #: unique identifier of the filter
        self.identifier = identifier
        #: user-defined name of the filter
        self.name = name
        #: whether or not to use this slot
        self.slot_used = True
        if identifier in Dataslot._instances:
            raise ValueError("Dataslot with identifier "
                             + "'{}' already exists!".format(identifier))
        Dataslot._instances[identifier] = self
        self.color = random_color()
        self.fl_name_dict = {"FL-1": "FL-1",
                             "FL-2": "FL-2",
                             "FL-3": "FL-3"}
        is_channel = cfg["setup"]["chip region"] == "channel"
        self.config = {
            "crosstalk": {
                # crosstalk
                "crosstalk fl12": 0,
                "crosstalk fl13": 0,
                "crosstalk fl21": 0,
                "crosstalk fl23": 0,
                "crosstalk fl31": 0,
                "crosstalk fl32": 0,
            },
            "emodulus": {
                # emodulus
                "emodulus enabled": is_channel,  # False for reservoir
                "emodulus model": "elastic sphere",
                "emodulus medium": "undefined",
                # https://dclab.readthedocs.io/en/latest/sec_av_emodulus.html
                # possible values are:
                # - "feature": scenario A
                # - None: secnario B ("emodulus medium" is "other" or user-def)
                # - "config": scenario C (temperature taken from config)
                # - "manual": scenario C (temperature entered manually)
                "emodulus scenario": "manual",
                "emodulus temperature": np.nan,
                "emodulus viscosity": np.nan,
            }
        }

        # use the emodulus medium and temperature values as defaults
        ds = self.get_dataset()
        calc = self.config["emodulus"]
        if "medium" in ds.config["setup"]:
            calc["emodulus medium"] = ds.config["setup"]["medium"]
        if "temp" in ds:
            # use the "temp" feature
            calc["emodulus scenario"] = "feature"
        elif "temperature" in ds.config["setup"]:
            # use the average temperature
            calc["emodulus temperature"] = ds.config["setup"]["temperature"]
            calc["emodulus scenario"] = "config"

        #: data file format
        self.format = self.get_dataset().format

    def __getstate__(self):
        state = {"color": self.color,
                 "crosstalk": self.config["crosstalk"],
                 "emodulus": self.config["emodulus"],
                 "fl names": self.fl_name_dict,
                 "identifier": self.identifier,
                 "name": self.name,
                 "path": self.path,
                 "slot used": self.slot_used,
                 }
        return copy.deepcopy(state)

    def __repr__(self):
        repre = "<Pipeline Slot '{}' at {}>".format(self.identifier,
                                                    hex(id(self)))
        return repre

    def __setstate__(self, state):
        if self.identifier != state["identifier"]:
            raise ValueError("Identifier mismatch: '{}' vs. '{}'".format(
                self.identifier, state["identifier"]))
        self.color = state["color"]
        self.config["crosstalk"].update(state["crosstalk"])
        self.config["emodulus"].update(state["emodulus"])
        self.fl_name_dict = state["fl names"]
        self.name = state["name"]
        self.path = state["path"]
        self.slot_used = state["slot used"]

    @staticmethod
    def get_slot(slot_id):
        """Get the slot with the given identifier.
        """
        return Dataslot._instances[slot_id]

    @staticmethod
    def get_instances():
        return Dataslot._instances

    @staticmethod
    def remove_slot(slot_id):
        """Remove a slot taking care of closing any opened files"""
        slot = Dataslot.get_slot(slot_id)
        ds = slot._dataset
        if ds is not None:
            if isinstance(ds, dclab.rtdc_dataset.RTDC_HDF5):
                ds._h5.close()
        Dataslot._instances.pop(slot_id)

    @property
    def hash(self):
        """Return the hash of the slot"""
        return hashobj(self.__getstate__())

    def _set_emodulus_config(self, dataset):
        """Set the Young's modulus data options

        The three cases in the dclab docs apply:
        https://dclab.readthedocs.io/en/latest/sec_av_emodulus.html
        """
        # remove any information
        for key in self.config["emodulus"]:
            if key in dataset.config["calculation"]:
                dataset.config["calculation"].pop(key)

        model = self.config["emodulus"]["emodulus model"]
        medium = self.config["emodulus"]["emodulus medium"]
        visc = self.config["emodulus"]["emodulus viscosity"]
        scenario = self.config["emodulus"]["emodulus scenario"]
        if scenario == "config":
            # Force the temperature from the dataset metadata
            # (this is a failsafe, the user/developer might not have set it)
            temp = dataset.config["setup"]["temperature"]
        elif scenario == "manual":
            # Only here do we actually need the temperature stored
            temp = self.config["emodulus"]["emodulus temperature"]
        else:
            # Temperature is not used in these scenarios.
            temp = np.nan

        dataset.config["calculation"]["emodulus model"] = model
        # known media
        if medium in dclab.features.emodulus.viscosity.KNOWN_MEDIA:
            dataset.config["calculation"]["emodulus medium"] = medium
        # temperature
        if not np.isnan(temp):
            dataset.config["calculation"]["emodulus temperature"] = temp
        # viscosity
        if medium not in KNOWN_MEDIA and not np.isnan(visc):
            dataset.config["calculation"]["emodulus viscosity"] = visc

    def get_dataset(self):
        """Return the corresponding dataset

        Returns
        -------
        ds: dclab.RTDCBase
            Loaded dataset
        """
        if self._dataset is None:
            ds = dclab.new_dataset(self.path, identifier=self.identifier)
            self._dataset = ds
        else:
            ds = self._dataset
        self.update_dataset(ds)
        return ds

    def update_dataset(self, dataset):
        """Update the configuration of an instance of RTDCBase

        This is used to update the configuration for computing
        the Young's modulus and fluorescence crosstalk.
        """
        # emodulus
        self._set_emodulus_config(dataset)

        # crosstalk
        if np.sum(list(self.config["crosstalk"].values())):
            dataset.config["calculation"].update(self.config["crosstalk"])
        else:
            # remove any information
            for key in self.config["crosstalk"]:
                if key in dataset.config["calculation"]:
                    dataset.config["calculation"].pop(key)


def random_color():
    color = "#"
    for _ in range(3):
        # dark colors (until 200)
        part = hex(np.random.randint(0, 200))[2:]
        if len(part) == 1:
            part += "0"
        color += part
    return color.upper()



import pkg_resources

from PyQt5 import uic, QtCore, QtWidgets

from ... import pipeline


class MatrixFilter(QtWidgets.QWidget):
    active_toggled = QtCore.pyqtSignal()
    enabled_toggled = QtCore.pyqtSignal(bool)
    option_action = QtCore.pyqtSignal(str)
    modify_clicked = QtCore.pyqtSignal(str)

    def __init__(self, identifier=None, state=None):
        QtWidgets.QWidget.__init__(self)
        path_ui = pkg_resources.resource_filename(
            "shapeout2.gui.matrix", "dm_filter.ui")
        uic.loadUi(path_ui, self)

        # options button
        menu = QtWidgets.QMenu()
        menu.addAction('duplicate', self.action_duplicate)
        menu.addAction('remove', self.action_remove)
        self.toolButton_opt.setMenu(menu)

        # toggle all active, all inactive, semi state
        self.toolButton_toggle.clicked.connect(self.active_toggled.emit)

        # toggle enabled/disabled state
        self.checkBox.clicked.connect(self.on_enabled_toggled)

        # modify filter button
        self.toolButton_modify.clicked.connect(self.on_modify)

        if state is None:
            filt = pipeline.Filter._instances[identifier]
            self.identifier = identifier
            self.name = filt.name
            # set tooltip/label
            self.update_content()
        else:
            self.__setstate__(state)

    def __getstate__(self):
        state = {"enabled": self.enabled,
                 "identifier": self.identifier,
                 "name": self.name,
                 }
        return state

    def __setstate__(self, state):
        if state["identifier"] not in pipeline.Filter._instances:
            # Create a new filter with the identifier
            pipeline.Filter(identifier=state["identifier"])
        self.identifier = state["identifier"]
        self.enabled = state["enabled"]
        self.name = state["name"]
        self.update_content()

    @property
    def enabled(self):
        filt = pipeline.Filter._instances[self.identifier]
        return filt.general["enable filters"]

    @enabled.setter
    def enabled(self, b):
        filt = pipeline.Filter._instances[self.identifier]
        filt.general["enable filters"] = b

    @property
    def name(self):
        filt = pipeline.Filter._instances[self.identifier]
        return filt.name

    @name.setter
    def name(self, text):
        filt = pipeline.Filter._instances[self.identifier]
        filt.name = text

    def action_duplicate(self):
        self.option_action.emit("duplicate")

    def action_remove(self):
        self.option_action.emit("remove")

    def on_enabled_toggled(self, b):
        self.enabled = b
        self.enabled_toggled.emit(b)

    def on_modify(self):
        self.modify_clicked.emit(self.identifier)

    def set_label_string(self, string):
        if self.label.fontMetrics().boundingRect(string).width() < 60:
            nstring = string
        else:
            nstring = string + "..."
            while True:
                width = self.label.fontMetrics().boundingRect(nstring).width()
                if width > 60:
                    nstring = nstring[:-4] + "..."
                else:
                    break
        self.label.setText(nstring)

    @QtCore.pyqtSlot()
    def update_content(self):
        """Reset tool tips and title"""
        self.label.setToolTip(self.name)
        self.set_label_string(self.name)
        self.checkBox.blockSignals(True)
        self.checkBox.setChecked(self.enabled)
        self.checkBox.blockSignals(False)
        self.enabled_toggled.emit(self.enabled)



from shapeout2 import util


def test_get_valid_filename():
    assert util.get_valid_filename("KLµ123$)]") == "KLu123?))"
    assert util.get_valid_filename("..KLµ123$)].") == "KLu123?))"
    assert util.get_valid_filename(". KLµ123$)]") == "_KLu123?))"


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



"""Test of Quick View set functionalities"""
import pathlib

from PyQt5 import QtCore

import dclab
import numpy as np
from shapeout2.gui.main import ShapeOut2
from shapeout2 import session
import pytest


datapath = pathlib.Path(__file__).parent / "data"


@pytest.fixture(autouse=True)
def run_around_tests():
    # Code that will run before your test, for example:
    session.clear_session()
    # A test function will be run at this point
    yield
    # Code that will run after your test, for example:
    session.clear_session()


def test_simple(qtbot):
    """Open the main window and close it again"""
    main_window = ShapeOut2()
    main_window.close()


def test_clear_session_issue_25(qtbot):
    """https://github.com/ZELLMECHANIK-DRESDEN/ShapeOut2/issues/25"""
    mw = ShapeOut2()
    qtbot.addWidget(mw)

    # add a dataslot
    path = pathlib.Path(__file__).parent / "data" / "calibration_beads_47.rtdc"
    mw.add_dataslot(paths=[path])

    assert len(mw.pipeline.slot_ids) == 1, "we added that"
    assert len(mw.pipeline.filter_ids) == 1, "automatically added"

    # activate a dataslot
    slot_id = mw.pipeline.slot_ids[0]
    filt_id = mw.pipeline.filter_ids[0]
    em = mw.block_matrix.get_widget(slot_id, filt_id)
    qtbot.mouseClick(em, QtCore.Qt.LeftButton, QtCore.Qt.ShiftModifier)
    # did that work?
    assert mw.toolButton_quick_view.isChecked()

    # now clear the session (this raised the errror in #25)
    mw.on_action_clear(assume_yes=True)


@pytest.mark.filterwarnings('ignore::RuntimeWarning')  # 0-div in kde-methods
@pytest.mark.filterwarnings('ignore::shapeout2.pipeline.core.'
                            + 'EmptyDatasetWarning')
def test_no_events_disable(qtbot):
    """When all events are removed, a message should be displayed"""
    mw = ShapeOut2()
    qtbot.addWidget(mw)

    # This session also contains a plot, so this is essentially
    # also a test for empty plots.
    spath = datapath / "version_2_1_6_no_events.so2"
    mw.on_action_open(spath)

    # Matrix widgets
    slot_id1 = mw.pipeline.slot_ids[0]
    slot_id2 = mw.pipeline.slot_ids[1]
    filt_id = mw.pipeline.filter_ids[0]
    em1 = mw.block_matrix.get_widget(slot_id1, filt_id)
    em2 = mw.block_matrix.get_widget(slot_id2, filt_id)

    # Now activate Quick View
    qtbot.mouseClick(em1, QtCore.Qt.LeftButton, QtCore.Qt.ShiftModifier)

    # Get Quick View instance
    qv = mw.widget_quick_view

    # This will display the "Hoppla!" message
    assert qv.label_noevents.isVisible()

    # Check the reverse
    qtbot.mouseClick(em2, QtCore.Qt.LeftButton, QtCore.Qt.ShiftModifier)
    assert not qv.label_noevents.isVisible()


def test_no_events_issue_37(qtbot):
    """https://github.com/ZELLMECHANIK-DRESDEN/ShapeOut2/issues/37

    "ValueError: zero-size array to reduction operation minimum
    which has no identity" when all events are filtered out.
    """
    mw = ShapeOut2()
    qtbot.addWidget(mw)

    # add a dataslot
    path = pathlib.Path(__file__).parent / "data" / "calibration_beads_47.rtdc"
    mw.add_dataslot(paths=[path, path])

    assert len(mw.pipeline.slot_ids) == 2, "we added those"
    assert len(mw.pipeline.filter_ids) == 1, "automatically added"

    # activate a dataslot
    slot_id = mw.pipeline.slot_ids[0]
    filt_id = mw.pipeline.filter_ids[0]
    em = mw.block_matrix.get_widget(slot_id, filt_id)
    qtbot.mouseClick(em, QtCore.Qt.LeftButton)  # activate
    # did that work?
    assert mw.pipeline.is_element_active(slot_id, filt_id)

    # filter away all events
    fe = mw.block_matrix.get_widget(filt_plot_id=filt_id)
    qtbot.mouseClick(fe.toolButton_modify, QtCore.Qt.LeftButton)
    fv = mw.widget_ana_view.widget_filter
    qtbot.mouseClick(fv.toolButton_moreless, QtCore.Qt.LeftButton)
    rc = fv._box_range_controls["area_um"]
    qtbot.mouseClick(rc.checkBox, QtCore.Qt.LeftButton)
    # did that work?
    assert rc.checkBox.isChecked()
    qtbot.mouseClick(fv.toolButton_moreless, QtCore.Qt.LeftButton)
    # set range
    rc.doubleSpinBox_min.setValue(0)
    rc.doubleSpinBox_max.setValue(1)
    qtbot.mouseClick(fv.pushButton_apply, QtCore.Qt.LeftButton)
    # did that work?
    ds = mw.pipeline.get_dataset(slot_index=0, filt_index=0,
                                 apply_filter=True)
    assert np.sum(ds.filter.all) == 0

    # open QuickView
    slot_id = mw.pipeline.slot_ids[0]
    filt_id = mw.pipeline.filter_ids[0]
    em = mw.block_matrix.get_widget(slot_id, filt_id)

    # this raised the error
    qtbot.mouseClick(em, QtCore.Qt.LeftButton, QtCore.Qt.ShiftModifier)
    mw.close()


def test_remove_dataset_h5py_error(qtbot):
    """Removing an activated dataset and activating Quick View fails

    Unhandled exception in Shape-Out version 2.0.1.post2:
    Traceback (most recent call last):
      File "/home/paul/repos/ShapeOut2/shapeout2/gui/main.py", line 235,
        in adopt_pipeline
        self.widget_quick_view.update_feature_choices()
      File "/home/paul/repos/ShapeOut2/shapeout2/gui/quick_view/qv_main.py",
        line 635, in update_feature_choices
        ds_feats = [f for f in self.rtdc_ds.features if f in feats_scalar]
    [...]
      File "/home/paul/repos/dclab/dclab/rtdc_dataset/fmt_hdf5.py", line 101,
        in _is_defective_feature
        if attr in self._h5.attrs:
    [...]
      File "h5py/_objects.pyx", line 55, in h5py._objects.with_phil.wrapper
      File "h5py/h5o.pyx", line 190, in h5py.h5o.open
    ValueError: Not a location (invalid object ID)
    """
    mw = ShapeOut2()
    qtbot.addWidget(mw)

    # add a dataslot
    path = pathlib.Path(__file__).parent / "data" / "calibration_beads_47.rtdc"
    mw.add_dataslot(paths=[path, path])

    assert len(mw.pipeline.slot_ids) == 2, "we added those"
    assert len(mw.pipeline.filter_ids) == 1, "automatically added"

    # activate a dataslot
    slot_id = mw.pipeline.slot_ids[0]
    filt_id = mw.pipeline.filter_ids[0]
    em = mw.block_matrix.get_widget(slot_id, filt_id)
    qtbot.mouseClick(em, QtCore.Qt.LeftButton)  # activate
    qtbot.mouseClick(em, QtCore.Qt.LeftButton, QtCore.Qt.ShiftModifier)
    # did that work?
    assert mw.pipeline.is_element_active(slot_id, filt_id)

    # close Quick View
    qtbot.mouseClick(mw.toolButton_quick_view, QtCore.Qt.LeftButton)

    # now remove the dataset
    pw = mw.block_matrix.get_widget(slot_id=slot_id)
    pw.action_remove()

    # open Quick View
    qtbot.mouseClick(mw.toolButton_quick_view, QtCore.Qt.LeftButton)
    mw.close()


def test_update_polygon_filter_issue_26(qtbot):
    """https://github.com/ZELLMECHANIK-DRESDEN/ShapeOut2/issues/26

    The recomputation of the filter ray was not triggered for some
    reason.
    """
    mw = ShapeOut2()
    qtbot.addWidget(mw)

    # add a dataslot
    path = pathlib.Path(__file__).parent / "data" / "calibration_beads_47.rtdc"
    filt_id = mw.add_filter()
    slot_ids = mw.add_dataslot(paths=[path])

    assert len(mw.pipeline.slot_ids) == 1, "we added that"
    assert len(mw.pipeline.filter_ids) == 1, "automatically added"

    # activate a dataslot
    slot_id = slot_ids[0]
    em = mw.block_matrix.get_widget(slot_id, filt_id)
    qtbot.mouseClick(em, QtCore.Qt.LeftButton, QtCore.Qt.ShiftModifier)

    em = mw.block_matrix.get_widget(slot_id, filt_id)
    qtbot.mouseClick(em, QtCore.Qt.LeftButton)

    # did that work?
    assert mw.toolButton_quick_view.isChecked()

    # Add a polygon filter
    assert len(dclab.PolygonFilter.instances) == 0
    qv = mw.widget_quick_view
    qtbot.mouseClick(qv.toolButton_poly, QtCore.Qt.LeftButton)
    qtbot.mouseClick(qv.pushButton_poly_create, QtCore.Qt.LeftButton)
    # three positions (not sure how to do this with mouse clicks)
    points = [[22, 0.01],
              [30, 0.01],
              [30, 0.014],
              ]
    qv.widget_scatter.set_poly_points(points)
    qtbot.mouseClick(qv.pushButton_poly_save, QtCore.Qt.LeftButton)
    # did that work?
    assert len(dclab.PolygonFilter.instances) == 1
    pf = dclab.PolygonFilter.instances[0]
    assert np.allclose(pf.points, points)

    # Add the polygon filter to the first filter
    fe = mw.block_matrix.get_widget(filt_plot_id=filt_id)
    qtbot.mouseClick(fe.toolButton_modify, QtCore.Qt.LeftButton)
    fv = mw.widget_ana_view.widget_filter
    cb = fv._polygon_checkboxes[pf.unique_id]
    qtbot.mouseClick(cb, QtCore.Qt.LeftButton)
    assert cb.isChecked()
    qtbot.mouseClick(fv.pushButton_apply, QtCore.Qt.LeftButton)
    # did that work?
    ds = mw.pipeline.get_dataset(slot_index=0, filt_index=0,
                                 apply_filter=True)
    assert np.sum(ds.filter.all) == 15

    # Modify the polygon filter
    qv.comboBox_poly.setCurrentIndex(1)
    points2 = [[22, 0.01],
               [30, 0.01],
               [30, 0.012],
               ]
    qv.widget_scatter.set_poly_points(points2)
    qtbot.mouseClick(qv.pushButton_poly_save, QtCore.Qt.LeftButton)
    assert len(dclab.PolygonFilter.instances) == 1
    pf2 = dclab.PolygonFilter.instances[0]
    assert np.allclose(pf2.points, points2)
    assert pf is pf2
    # now the filter should be updated (this worked already)
    ds2 = mw.pipeline.get_dataset(slot_index=0, filt_index=0,
                                  apply_filter=True)
    assert np.sum(ds2.filter.all) == 8
    # but the plots were not updated in #26
    assert len(qv.widget_scatter.scatter.getData()[0]) == 8



"""Test computation of emodulus"""
import pathlib
import tempfile

from PyQt5 import QtCore, QtWidgets

import dclab
import h5py
import numpy as np
from shapeout2.gui.main import ShapeOut2
from shapeout2 import session
import pytest

datapath = pathlib.Path(__file__).parent / "data"


def make_dataset(medium="CellCarrier", temp=22.5, temp_range=[22, 23],
                 chip_region="channel"):
    # create a fake dataset
    path = datapath / "calibration_beads_47.rtdc"
    ds = dclab.new_dataset(path)
    tmp = tempfile.mktemp(".rtdc", prefix="example_")
    ds.export.hdf5(tmp, features=["deform", "area_um", "bright_avg"])
    with h5py.File(tmp, mode="a") as h5:
        h5["events/temp"] = np.linspace(temp_range[0], temp_range[1], len(ds))
        if medium is None:
            h5.attrs.pop("setup:medium")
        else:
            h5.attrs["setup:medium"] = medium
        h5.attrs["setup:temperature"] = temp
        h5.attrs["setup:chip region"] = chip_region
    return pathlib.Path(tmp)


@pytest.fixture(autouse=True)
def run_around_tests():
    # Code that will run before your test, for example:
    session.clear_session()
    # A test function will be run at this point
    yield
    # Code that will run after your test, for example:
    session.clear_session()


def test_allow_to_set_manual_temperature_for_known_medium(qtbot):
    """Fixes regression introduced in 2.4.0"""
    mw = ShapeOut2()
    qtbot.addWidget(mw)
    # add fake measurement
    path1 = make_dataset(medium=None, temp=23.5)
    mw.add_dataslot(paths=[path1])
    wsl = mw.widget_ana_view.widget_slot

    # 1. test whether we can actually select things in comboBox_temp
    ccidx = wsl.comboBox_medium.findData("CellCarrier")
    wsl.comboBox_medium.setCurrentIndex(ccidx)
    assert wsl.comboBox_temp.isEnabled()

    # 2. test whether we can select manual
    manidx = wsl.comboBox_temp.findData("manual")
    wsl.comboBox_temp.setCurrentIndex(manidx)
    assert wsl.doubleSpinBox_temp.isEnabled()
    assert not wsl.doubleSpinBox_temp.isReadOnly()

    # 3. test whether we can select config and the the temperature
    # should be 23.5
    conidx = wsl.comboBox_temp.findData("config")
    wsl.comboBox_temp.setCurrentIndex(conidx)
    assert wsl.doubleSpinBox_temp.isEnabled()
    assert wsl.doubleSpinBox_temp.isReadOnly()
    assert wsl.doubleSpinBox_temp.value() == 23.5


def test_empty_medium_string_should_offer_user_edit(qtbot):
    mw = ShapeOut2()
    qtbot.addWidget(mw)
    # add fake measurement
    path1 = make_dataset(medium=" ")
    mw.add_dataslot(paths=[path1])
    wsl = mw.widget_ana_view.widget_slot
    ds = mw.pipeline.slots[0].get_dataset()
    assert ds.config["setup"]["medium"] == " "
    assert wsl.comboBox_medium.currentData() == "other"
    assert not wsl.doubleSpinBox_visc.isReadOnly(), "Should be editable"
    assert wsl.doubleSpinBox_visc.isEnabled(), "Should be editable"
    assert np.isnan(wsl.__getstate__()["emodulus"]["emodulus temperature"])
    assert wsl.__getstate__()["emodulus"]["emodulus medium"] == "other"
    assert wsl.__getstate__()["emodulus"]["emodulus scenario"] is None


def test_other_medium_viscosity_editable_issue_49(qtbot):
    mw = ShapeOut2()
    qtbot.addWidget(mw)
    # add fake measurement
    path1 = make_dataset(medium=None)
    mw.add_dataslot(paths=[path1])
    wsl = mw.widget_ana_view.widget_slot
    ds = mw.pipeline.slots[0].get_dataset()
    assert "medium" not in ds.config["setup"], "sanity check (medium removed)"
    oidx = wsl.comboBox_medium.findData("other")
    wsl.comboBox_medium.setCurrentIndex(oidx)
    assert not wsl.doubleSpinBox_visc.isReadOnly(), "Should be editable"
    assert wsl.doubleSpinBox_visc.isEnabled(), "Should be editable"


@pytest.mark.filterwarnings(
    'ignore::dclab.features.emodulus.YoungsModulusLookupTableExceededWarning')
def test_simple(qtbot):
    mw = ShapeOut2()
    qtbot.addWidget(mw)

    # add fake measurement
    path = make_dataset(medium="CellCarrier", temp=22.5, temp_range=[22, 23])

    with dclab.new_dataset(path) as ds:
        # Youngs modulus can readily be computed
        # https://dclab.readthedocs.io/en/latest/sec_av_emodulus.html
        ds.config["calculation"]["emodulus medium"] = \
            ds.config["setup"]["medium"]
        ds.config["calculation"]["emodulus model"] = "elastic sphere"
        emodA = np.array(ds["emodulus"], copy=True)
        ds.config["calculation"]["emodulus temperature"] = \
            ds.config["setup"]["temperature"]
        emodC = np.array(ds["emodulus"], copy=True)
        assert not np.allclose(emodA, emodC, atol=0, rtol=1e-12,
                               equal_nan=True), "sanity check"

    mw.add_dataslot(paths=[path])
    wsl = mw.widget_ana_view.widget_slot

    # default values
    assert wsl.comboBox_medium.currentData() == "CellCarrier"
    assert wsl.comboBox_temp.currentData() == "feature"

    # scenario A (this is already set by default)
    ds1 = mw.pipeline.slots[0].get_dataset()
    assert np.allclose(ds1["emodulus"], emodA, atol=0, rtol=1e-12,
                       equal_nan=True)

    # scenario C (switch to config)
    wsl.comboBox_temp.setCurrentIndex(wsl.comboBox_temp.findData("config"))
    qtbot.mouseClick(wsl.pushButton_apply, QtCore.Qt.LeftButton)
    ds2 = mw.pipeline.slots[0].get_dataset()
    assert np.allclose(ds2["emodulus"], emodC, atol=0, rtol=1e-12,
                       equal_nan=True)

    # scenario C (switch to manual)
    wsl.comboBox_temp.setCurrentIndex(wsl.comboBox_temp.findData("manual"))
    wsl.doubleSpinBox_temp.setValue(22.5)
    qtbot.mouseClick(wsl.pushButton_apply, QtCore.Qt.LeftButton)
    ds3 = mw.pipeline.slots[0].get_dataset()
    assert np.allclose(ds3["emodulus"], emodC, atol=0, rtol=1e-12,
                       equal_nan=True)

    try:
        path.unlink()
    except BaseException:
        pass


def test_switch_and_update_chip_region(qtbot):
    mw = ShapeOut2()
    qtbot.addWidget(mw)

    # add fake measurement
    path1 = make_dataset(medium="CellCarrier", temp=22.5, temp_range=[22, 23],
                         chip_region="channel")
    path2 = make_dataset(medium="CellCarrier", temp=22.5, temp_range=[22, 23],
                         chip_region="reservoir")

    slot_id1, slot_id2 = mw.add_dataslot(paths=[path1, path2])
    wsl = mw.widget_ana_view.widget_slot

    # select the first slot
    em1 = mw.block_matrix.get_widget(slot_id=slot_id1)
    em2 = mw.block_matrix.get_widget(slot_id=slot_id2)
    qtbot.mouseClick(em1.toolButton_modify, QtCore.Qt.LeftButton)
    # set temperature manually
    idm = wsl.comboBox_temp.findData("manual")
    wsl.comboBox_temp.setCurrentIndex(idm)
    assert wsl.comboBox_temp.currentData() == "manual"
    wsl.doubleSpinBox_temp.setValue(20.0)
    assert wsl.doubleSpinBox_temp.value() == 20
    qtbot.mouseClick(wsl.pushButton_apply, QtCore.Qt.LeftButton)
    QtWidgets.QApplication.processEvents()
    # check whether that worked
    assert wsl.get_dataset(
    ).config["calculation"]["emodulus temperature"] == 20

    # switch to the second (reservoir) measurement
    qtbot.mouseClick(em2.toolButton_modify, QtCore.Qt.LeftButton)
    assert not wsl.groupBox_emod.isVisible()
    # now switch back
    qtbot.mouseClick(em1.toolButton_modify, QtCore.Qt.LeftButton)

    # This is the actual test
    assert wsl.doubleSpinBox_temp.value() == 20

    try:
        path1.unlink()
        path2.unlink()
    except BaseException:
        pass


def test_switch_and_update_medium(qtbot):
    mw = ShapeOut2()
    qtbot.addWidget(mw)

    # add fake measurement
    path1 = make_dataset(medium="CellCarrier", temp=22.5, temp_range=[22, 23])
    path2 = make_dataset(medium="UserDefined", temp=22.5, temp_range=[22, 23])

    slot_id1, slot_id2 = mw.add_dataslot(paths=[path1, path2])
    wsl = mw.widget_ana_view.widget_slot

    # select the first slot
    em1 = mw.block_matrix.get_widget(slot_id=slot_id1)
    em2 = mw.block_matrix.get_widget(slot_id=slot_id2)
    qtbot.mouseClick(em1.toolButton_modify, QtCore.Qt.LeftButton)
    # set temperature manually
    idm = wsl.comboBox_temp.findData("manual")
    wsl.comboBox_temp.setCurrentIndex(idm)
    assert wsl.comboBox_temp.currentData() == "manual"
    wsl.doubleSpinBox_temp.setValue(20.0)
    assert wsl.doubleSpinBox_temp.value() == 20
    qtbot.mouseClick(wsl.pushButton_apply, QtCore.Qt.LeftButton)
    QtWidgets.QApplication.processEvents()
    # check whether that worked
    assert wsl.get_dataset(
    ).config["calculation"]["emodulus temperature"] == 20

    # switch to the second (user-defined medium) measurement
    qtbot.mouseClick(em2.toolButton_modify, QtCore.Qt.LeftButton)
    assert wsl.comboBox_medium.currentData() == "UserDefined"
    assert not wsl.doubleSpinBox_visc.isReadOnly(), "Should be editable"
    assert wsl.doubleSpinBox_visc.isEnabled(), "Should be editable"
    assert not wsl.doubleSpinBox_temp.isEnabled(), "Should not be editable"
    # now switch back
    qtbot.mouseClick(em1.toolButton_modify, QtCore.Qt.LeftButton)

    # This is the actual test
    assert wsl.doubleSpinBox_temp.value() == 20
    assert wsl.doubleSpinBox_temp.isEnabled(), "Should be editable"

    try:
        path1.unlink()
        path2.unlink()
    except BaseException:
        pass


def test_user_defined_medium_should_work(qtbot):
    mw = ShapeOut2()
    qtbot.addWidget(mw)
    # add fake measurement
    path1 = make_dataset(medium="MyMedium")
    mw.add_dataslot(paths=[path1])
    wsl = mw.widget_ana_view.widget_slot
    ds = mw.pipeline.slots[0].get_dataset()
    assert ds.config["setup"]["medium"] == "MyMedium", "sanity check"
    assert wsl.comboBox_medium.currentData() == "MyMedium"
    assert not wsl.doubleSpinBox_visc.isReadOnly(), "Should be editable"
    assert wsl.doubleSpinBox_visc.isEnabled(), "Should be editable"
    wsl.doubleSpinBox_visc.setValue(12.1)
    assert wsl.__getstate__()["emodulus"]["emodulus viscosity"] == 12.1
    assert np.isnan(wsl.__getstate__()["emodulus"]["emodulus temperature"])
    assert wsl.__getstate__()["emodulus"]["emodulus medium"] == "MyMedium"
    assert wsl.__getstate__()["emodulus"]["emodulus scenario"] is None



"""Test plot export"""
import os
import pathlib
import shutil
import tempfile

from PyQt5 import QtWidgets
import pytest
from shapeout2.gui.main import ShapeOut2
from shapeout2.gui import export
from shapeout2 import session


datapath = pathlib.Path(__file__).parent / "data"


@pytest.fixture(autouse=True)
def run_around_tests():
    # Code that will run before your test, for example:
    session.clear_session()
    # A test function will be run at this point
    yield
    # Code that will run after your test, for example:
    session.clear_session()


def test_export_single_plot(qtbot, monkeypatch):
    """Export of single plots not possible up until version 2.1.4"""
    spath = datapath / "version_2_1_0_basic.so2"

    mw = ShapeOut2()
    qtbot.addWidget(mw)

    mw.on_action_open(spath)

    # perform the export
    tmpd = tempfile.mkdtemp(suffix="", prefix="shapeout2_test_plot_export_")

    tmpf = os.path.join(tmpd, "no_suffix")
    assert not pathlib.Path(tmpf).exists()
    monkeypatch.setattr(QtWidgets.QFileDialog, "getSaveFileName",
                        lambda *args: (tmpf, ""))

    # create export dialog manually
    dlg = export.ExportPlot(mw, pipeline=mw.pipeline)

    # select a single plot to export
    plot_id = mw.pipeline.plot_ids[0]
    assert isinstance(plot_id, str)
    plot_index = dlg.comboBox_plot.findData(plot_id)
    assert plot_index > 0
    dlg.comboBox_plot.setCurrentIndex(plot_index)
    assert dlg.comboBox_plot.currentData() == plot_id

    dlg.export_plots()
    assert pathlib.Path(tmpf).with_suffix(".png").exists()

    # cleanup
    shutil.rmtree(tmpd, ignore_errors=True)



from dclab.rtdc_dataset import ancillaries

#: integer-valued features
INTEGER_FEATURES = [
    "fl1_max",
    "fl1_npeaks",
    "fl2_max",
    "fl2_npeaks",
    "fl3_max",
    "fl3_npeaks",
    "frame",
    "index",
    "ml_class",
    "nevents",
]

#: ancillary features that are easily computed
QUICK_FEATURES = ancillaries.FEATURES_RAPID



import numpy as np
from PyQt5 import QtGui
import pyqtgraph as pg
from pyqtgraph.graphicsItems.GradientEditorItem import Gradients


class ColorBarWidget(pg.GraphicsWidget):
    def __init__(self, cmap, width, height, vmin=0.0, vmax=1.0, label=""):
        """Colorbar widget that can be added to a layout

        This widget can be added to a GraphicsLayout. It is designed to
        work well with the Shape-Out plot layout.

        Parameters
        ----------
        cmap: str
            Name of the colormap
        width: int
            Width of the colorbar
        height:
            Height of the colorbar
        vmin: float
            Lower value of the colorbar
        vmax: float
            Upper value of the colorbar
        label: str
            Label placed next to the colorbar

        Notes
        -----
        Inspired by https://gist.github.com/maedoc/b61090021d2a5161c5b9
        """
        pg.GraphicsWidget.__init__(self)
        self.setSizePolicy(QtGui.QSizePolicy.Fixed,
                           QtGui.QSizePolicy.Preferred)

        # arguments
        pcmap = pg.ColorMap(*zip(*Gradients[cmap]["ticks"]))
        w = width
        h = height
        stops, colors = pcmap.getStops('float')
        smn, spp = stops.min(), stops.ptp()
        stops = (stops - stops.min())/stops.ptp()
        ticks = np.r_[0.0:1.0:5j, 1.0] * spp + smn
        tick_labels = ["%0.2g" % (t,) for t in np.linspace(vmin, vmax, 5)]

        # setup picture
        self.pic = pg.QtGui.QPicture()
        p = pg.QtGui.QPainter(self.pic)

        # draw bar with gradient following colormap
        p.setPen(pg.mkPen('k'))
        grad = pg.QtGui.QLinearGradient(w/2.0, 0.0, w/2.0, h*1.0)
        for stop, color in zip(stops, colors):
            grad.setColorAt(1.0 - stop, pg.QtGui.QColor(*[c for c in color]))
        p.setBrush(pg.QtGui.QBrush(grad))
        p.drawRect(pg.QtCore.QRectF(0, 0, w, h))

        # draw ticks & tick labels
        mintx = 0.0
        maxwidth = 0.0
        for tick, tick_label in zip(ticks, tick_labels):
            y_ = (1.0 - (tick - smn)/spp) * h
            p.drawLine(w, int(y_), int(w+5.0), int(y_))
            br = p.boundingRect(
                0, 0, 0, 0, pg.QtCore.Qt.AlignRight, tick_label)
            if br.x() < mintx:
                mintx = br.x()
            if br.width() > maxwidth:
                maxwidth = br.width()
            p.drawText(int(br.x() + 10.0 + w + br.width()),
                       int(y_ + br.height() / 4.0),
                       tick_label)

        # draw label
        br = p.boundingRect(0, 0, 0, 0, pg.QtCore.Qt.AlignBottom, label)
        p.rotate(90)
        p.drawText(int(h/2 - br.width()/2),
                   int(-w-maxwidth-15),
                   label)

        # done
        p.end()

        # set minimum sizes (how do you get the actual bounding rect?)
        br = self.pic.boundingRect()
        self.setMinimumWidth(br.width() + maxwidth + 20)
        self.setMinimumHeight(h)

        # alognment with other Shape-Out plots (kind of a workaround)
        self.translate(0, 40)

    def paint(self, p, *args):
        # paint underlying mask
        p.setPen(pg.QtGui.QColor(255, 255, 255, 0))
        p.setBrush(pg.QtGui.QColor(255, 255, 255, 200))

        # paint colorbar
        p.drawPicture(0, 0, self.pic)



from . import dm_element


class MatrixElement(dm_element.MatrixElement):
    def mousePressEvent(self, event):
        # toggle selection
        if not self.invalid:
            self.active = not self.active
            self.element_changed.emit()
            self.update_content()
            event.accept()

    def update_content(self, quickview=False):
        if self.invalid:
            color = "#DCDCDC"  # gray
            label = "invalid"
            tooltip = "Incompatible plot settings"
        elif self.active and self.enabled:
            color = "#86E7C1"  # turquois
            label = "active"
            tooltip = "Click to deactivate"
        elif self.active and not self.enabled:
            color = "#C9DAD7"  # gray-turquois
            label = "active\n(unused)"
            tooltip = "Click to deactivate"
        elif not self.active and self.enabled:
            color = "#EFEFEF"  # light gray
            label = "inactive"
            tooltip = "Click to activate"
        else:
            color = "#DCDCDC"  # gray
            label = "inactive\n(unused)"
            tooltip = "Click to activate"

        self.setStyleSheet("background-color:{}".format(color))
        self.label.setStyleSheet("background-color:{}".format(color))
        self.label.setText(label)
        self.setToolTip(tooltip)
        self.label.setToolTip(tooltip)



import numpy as np
from PyQt5 import QtCore, QtGui, QtWidgets
import pyqtgraph as pg

from .simple_plot_widget import SimpleViewBox


class SimpleImageView(pg.ImageView):
    """Custom class for data visualization in Shape-Out"""

    def __init__(self, *args, **kwargs):
        super(SimpleImageView, self).__init__(view=SimpleImageViewBox(),
                                              *args, **kwargs)
        self.view.export.connect(self.on_export)

        # disable pyqtgraph controls we don't need
        self.ui.histogram.hide()
        self.ui.roiBtn.hide()
        self.ui.menuBtn.hide()
        # disable keyboard shortcuts
        self.keyPressEvent = lambda _: None
        self.keyReleaseEvent = lambda _: None

    def on_export(self, suffix):
        assert suffix == "png"
        path, _ = QtWidgets.QFileDialog.getSaveFileName(
            self, '', '', 'PNG image (*.png)', '')
        if not path.endswith(".png"):
            path += ".png"

        img = np.require(self.image, np.uint8, 'C')
        height, width, _ = self.image.shape

        qImg = QtGui.QImage(img, width, height, width *
                            3, QtGui.QImage.Format_RGB888)
        qImg.save(path)


class SimpleImageViewBox(SimpleViewBox):
    export = QtCore.pyqtSignal(str)

    def raiseContextMenu(self, ev):
        menu = self.menu
        menu.clear()
        menu.addAction("Save event image as PNG",
                       lambda: self.export.emit("png"))

        pos = ev.screenPos()
        menu.popup(QtCore.QPoint(pos.x(), pos.y()))
        return True



"""Plotting GUI tests"""
import pathlib
import tempfile

import dclab
import h5py
import numpy as np
from PyQt5 import QtCore
import pytest
from shapeout2.gui.main import ShapeOut2
from shapeout2 import pipeline, session


datapath = pathlib.Path(__file__).parent / "data"


@pytest.fixture(autouse=True)
def run_around_tests():
    # Code that will run before your test, for example:
    session.clear_session()
    # A test function will be run at this point
    yield
    # Code that will run after your test, for example:
    session.clear_session()


def test_empty_plot_with_one_plot_per_dataset_issue_41(qtbot):
    """
    Setting "one plot per dataset" for an empty plot resulted in
    zero-division error when determining col/row numbers
    """
    mw = ShapeOut2()
    qtbot.addWidget(mw)

    # add a dataslot
    path = datapath / "calibration_beads_47.rtdc"
    mw.add_dataslot(paths=[path])

    # add a plot
    plot_id = mw.add_plot()

    # activate analysis view
    pe = mw.block_matrix.get_widget(filt_plot_id=plot_id)
    qtbot.mouseClick(pe.toolButton_modify, QtCore.Qt.LeftButton)

    pv = mw.widget_ana_view.widget_plot

    # Change to "each" and apply
    idx = pv.comboBox_division.findData("each")
    pv.comboBox_division.setCurrentIndex(idx)
    # Lead to zero-division error in "get_plot_col_row_count"
    qtbot.mouseClick(pv.pushButton_apply, QtCore.Qt.LeftButton)


def test_feature_bright_avg_not_present_issue_62(qtbot):
    """Plot a dataset that does not contain the "bright_avg" feature

    ...or any means of computing it (i.e. via "image")
    """
    # create fake dataset without bright_avg
    tmp = tempfile.mktemp(".rtdc", prefix="example_hue_")
    with dclab.new_dataset(datapath / "calibration_beads_47.rtdc") as ds:
        ds.export.hdf5(tmp, features=["area_um", "pos_x", "pos_y", "deform"])

    mw = ShapeOut2()
    qtbot.addWidget(mw)
    # add dataset
    slot_id = mw.add_dataslot([tmp])[0]
    # add plot
    plot_id = mw.add_plot()
    # and activate it
    pw = mw.block_matrix.get_widget(filt_plot_id=plot_id, slot_id=slot_id)
    # this raised "ValueError: 'bright_avg' is not in list" (issue #62)
    qtbot.mouseClick(pw, QtCore.Qt.LeftButton)


def test_handle_axis_selection_empty_plot(qtbot):
    """User did not add a dataset to a plot and starts changing plot params"""
    mw = ShapeOut2()
    qtbot.addWidget(mw)

    # add a dataslot
    path = datapath / "calibration_beads_47.rtdc"
    mw.add_dataslot(paths=[path])

    # add a plot
    plot_id = mw.add_plot()

    assert len(mw.pipeline.slot_ids) == 1, "we added that"
    assert len(mw.pipeline.filter_ids) == 1, "automatically added"
    assert len(mw.pipeline.plot_ids) == 1, "we added that"

    # activate analysis view
    pe = mw.block_matrix.get_widget(filt_plot_id=plot_id)
    qtbot.mouseClick(pe.toolButton_modify, QtCore.Qt.LeftButton)

    pv = mw.widget_ana_view.widget_plot

    # This lead to:
    #    Traceback (most recent call last):
    #  File "/ShapeOut2/shapeout2/gui/analysis/ana_plot.py",
    #     line 406, in on_axis_changed
    #    self._set_contour_spacing_auto(axis_y=gen["axis y"])
    #  File "/ShapeOut2/shapeout2/gui/analysis/ana_plot.py",
    #     line 361, in _set_contour_spacing_auto
    #    spacings_xy.append(np.min(spacings))
    #  File "/numpy/core/fromnumeric.py", line 2618, in amin
    #    initial=initial)
    #  File "/numpy/core/fromnumeric.py", line 86, in _wrapreduction
    #    return ufunc.reduce(obj, axis, dtype, out, **passkwargs)
    # ValueError: zero-size array to reduction operation minimum which
    # has no identity
    pv.comboBox_axis_y.setCurrentIndex(pv.comboBox_axis_y.findData("emodulus"))


def test_handle_empty_plots_issue_27(qtbot):
    """Correctly handle empty plots

    https://github.com/ZELLMECHANIK-DRESDEN/ShapeOut2/issues/27
    """
    mw = ShapeOut2()
    qtbot.addWidget(mw)

    # add a dataslot
    path = datapath / "calibration_beads_47.rtdc"
    mw.add_dataslot(paths=[path])
    # add another one
    mw.add_dataslot(paths=[path])

    assert len(mw.pipeline.slot_ids) == 2, "we added those"
    assert len(mw.pipeline.filter_ids) == 1, "automatically added"

    # activate a dataslot
    slot_id = mw.pipeline.slot_ids[0]
    filt_id = mw.pipeline.filter_ids[0]
    em = mw.block_matrix.get_widget(slot_id, filt_id)
    qtbot.mouseClick(em, QtCore.Qt.LeftButton)  # activate
    # did that work?
    assert mw.pipeline.is_element_active(slot_id, filt_id)

    # filter away all events
    fe = mw.block_matrix.get_widget(filt_plot_id=filt_id)
    qtbot.mouseClick(fe.toolButton_modify, QtCore.Qt.LeftButton)
    fv = mw.widget_ana_view.widget_filter
    qtbot.mouseClick(fv.toolButton_moreless, QtCore.Qt.LeftButton)
    rc = fv._box_range_controls["area_um"]
    qtbot.mouseClick(rc.checkBox, QtCore.Qt.LeftButton)
    # did that work?
    assert rc.checkBox.isChecked()
    qtbot.mouseClick(fv.toolButton_moreless, QtCore.Qt.LeftButton)
    # set range
    rc.doubleSpinBox_min.setValue(0)
    rc.doubleSpinBox_max.setValue(1)
    qtbot.mouseClick(fv.pushButton_apply, QtCore.Qt.LeftButton)
    # did that work?
    ds = mw.pipeline.get_dataset(slot_index=0, filt_index=0,
                                 apply_filter=True)
    assert np.sum(ds.filter.all) == 0

    # now create a plot window
    plot_id = mw.add_plot()
    pe = mw.block_matrix.get_widget(slot_id, plot_id)
    with pytest.warns(pipeline.core.EmptyDatasetWarning):
        # this now only throws a warning
        qtbot.mouseClick(pe, QtCore.Qt.LeftButton)  # activate (raises #27)


@pytest.mark.filterwarnings(
    'ignore::dclab.features.emodulus.YoungsModulusLookupTableExceededWarning')
def test_handle_nan_valued_feature_color(qtbot):
    """User wants to color scatter data points with feature containing nans"""
    spath = datapath / "version_2_1_2_plot_color_emodulus.so2"

    mw = ShapeOut2()
    qtbot.addWidget(mw)

    # lead to:
    # OverflowError: argument 4 overflowed: value must be in the range
    # -2147483648 to 2147483647
    mw.on_action_open(spath)


def test_hue_feature_not_computed_if_not_selected(qtbot):
    # generate .rtdc file without bright_avg feature
    tmp = tempfile.mktemp(".rtdc", prefix="example_hue_")
    with dclab.new_dataset(datapath / "calibration_beads_47.rtdc") as ds:
        ds.export.hdf5(tmp, features=["area_um", "pos_x", "pos_y", "image",
                                      "mask", "deform"])
    mw = ShapeOut2()
    qtbot.addWidget(mw)
    # add dataset
    slot_id = mw.add_dataslot([tmp])[0]
    # add plot
    plot_id = mw.add_plot()
    # and activate it
    pw = mw.block_matrix.get_widget(filt_plot_id=plot_id, slot_id=slot_id)
    qtbot.mouseClick(pw, QtCore.Qt.LeftButton)
    # get the dataset
    ds = mw.pipeline.get_dataset(slot_index=0)
    # check whether the item has been plotted
    datasets, _ = mw.pipeline.get_plot_datasets(plot_id)
    assert datasets[0] is ds
    # now check whether "bright_avg" has been computed
    assert "bright_avg" in ds.features
    assert "bright_avg" not in ds.features_loaded


def test_plot_ml_score(qtbot):
    tmp = tempfile.mktemp(".rtdc", prefix="example_ml_score_")
    with dclab.new_dataset(datapath / "calibration_beads_47.rtdc") as ds:
        ds.export.hdf5(tmp, features=["area_um", "pos_x", "pos_y", "image",
                                      "mask", "deform"])
        lends = len(ds)
    # add ml_score features
    with h5py.File(tmp, "a") as h5:
        h5["/events/ml_score_ds9"] = np.linspace(0, 1, lends)
        h5["/events/ml_score_voy"] = np.linspace(1, 0, lends)
    mw = ShapeOut2()
    qtbot.addWidget(mw)
    # add dataset
    slot_id = mw.add_dataslot([tmp])[0]
    # add plot
    plot_id = mw.add_plot()
    # and activate it
    pw = mw.block_matrix.get_widget(filt_plot_id=plot_id, slot_id=slot_id)
    qtbot.mouseClick(pw, QtCore.Qt.LeftButton)
    # get the dataset
    ds = mw.pipeline.get_dataset(slot_index=0)
    # sanity check
    assert "ml_class" in ds

    # Now set the x axis to Voyager
    pv = mw.widget_ana_view.widget_plot
    idvoy = pv.comboBox_axis_x.findData("ml_score_voy")
    assert idvoy >= 0
    pv.comboBox_axis_x.setCurrentIndex(idvoy)
    qtbot.mouseClick(pv.pushButton_apply, QtCore.Qt.LeftButton)

    try:
        pathlib.Path(tmp).unlink()
    except OSError:
        pass


def test_remove_plots_issue_36(qtbot):
    """Correctly handle empty plots

    https://github.com/ZELLMECHANIK-DRESDEN/ShapeOut2/issues/36

    Traceback (most recent call last):
      File "/home/paul/repos/ShapeOut2/shapeout2/gui/main.py",
        line 193, in adopt_pipeline
        lay = pipeline_state["plots"][plot_index]["layout"]
    IndexError: list index out of range
    """
    mw = ShapeOut2()
    qtbot.addWidget(mw)

    # add a dataslots
    path = datapath / "calibration_beads_47.rtdc"
    mw.add_dataslot(paths=[path, path, path])

    assert len(mw.pipeline.slot_ids) == 3, "we added those"
    assert len(mw.pipeline.filter_ids) == 1, "automatically added"

    # now create a plot window
    plot_id = mw.add_plot()
    # and another one
    mw.add_plot()

    # remove a plot
    pw = mw.block_matrix.get_widget(filt_plot_id=plot_id)
    pw.action_remove()
